import warnings

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from xeofs.utils.data_types import DataArray, DataList, DataSet

warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")


# =============================================================================
# Synthetic data
# =============================================================================
def generate_synthetic_dataarray(
    n_sample=1,
    n_feature=1,
    index_policy="index",
    nan_policy="no_nan",
    dask_policy="no_dask",
    seed=0,
) -> DataArray:
    """Create synthetic DataArray.

    Parameters:
    ------------
    n_sample: int
        Number of sample dimensions.
    n_dims_feature: int
        Number of feature dimensions.
    index_policy: ["index", "multiindex"], default="index"
        If "multiindex", the data will have a multiindex.
    nan_policy: ["no_nan", "isolated", "fulldim"], default="no_nan"
        If specified, the data will contain NaNs.
    dsak_policy: ["no_dask", "dask"], default="no_dask"
        If "dask", the data will be a dask array.
    seed: int, default=0
        Seed for the random number generator.

    Returns:
    ---------
    data: xr.DataArray
        Synthetic data.

    """
    rng = np.random.default_rng(seed)

    # Create dimensions
    sample_dims = [f"sample{i}" for i in range(n_sample)]
    feature_dims = [f"feature{i}" for i in range(n_feature)]
    all_dims = sample_dims + feature_dims

    # Create coordinates/indices
    coords = {}
    for i, dim in enumerate(all_dims):
        if index_policy == "multiindex":
            coords[dim] = pd.MultiIndex.from_arrays(
                [np.arange(6 - i), np.arange(6 - i)],
                names=[f"index{i}a", f"index{i}b"],
            )
        elif index_policy == "index":
            coords[dim] = np.arange(6 + i)
        else:
            raise ValueError(f"Invalid value for index_policy: {index_policy}")

    # Get data shape
    shape = tuple([len(coords[dim]) for dim in all_dims])

    # Create data
    noise = rng.normal(5, 3, size=shape)
    signal = 2 * np.sin(np.linspace(0, 2 * np.pi, shape[-1]))
    signal = np.broadcast_to(signal, shape)
    data = signal + noise
    data = xr.DataArray(data, dims=all_dims, coords=coords)

    # Add NaNs
    if nan_policy == "no_nan":
        pass
    elif nan_policy == "isolated":
        isolated_point = {dim: 0 for dim in all_dims}
        data.loc[isolated_point] = np.nan
    elif nan_policy == "fulldim":
        fulldim_point = {dim: 0 for dim in feature_dims}
        data.loc[fulldim_point] = np.nan
    else:
        raise ValueError(f"Invalid value for nan_policy: {nan_policy}")

    # Convert to dask array
    if dask_policy == "no_dask":
        pass
    elif dask_policy == "dask":
        data = data.chunk({"sample0": 1})
    else:
        raise ValueError(f"Invalid value for dask_policy: {dask_policy}")

    return data


def generate_synthetic_dataset(
    n_variables=1,
    n_sample=1,
    n_feature=1,
    index_policy="index",
    nan_policy="no_nan",
    dask_policy="no_dask",
    seed=0,
) -> DataSet:
    """Create synthetic Dataset.

    Parameters:
    ------------
    n_variables: int
        Number of variables.
    n_sample: int
        Number of sample dimensions.
    n_dims_feature: int
        Number of feature dimensions.
    index_policy: ["index", "multiindex"], default="index"
        If "multiindex", the data will have a multiindex.
    nan_policy: ["no_nan", "isolated", "fulldim"], default="no_nan"
        If specified, the data will contain NaNs.
    dask_policy: ["no_dask", "dask"], default="no_dask"
        If "dask", the data will be a dask array.
    seed: int, default=0
        Seed for the random number generator.

    Returns:
    ---------
    data: xr.Dataset
        Synthetic data.

    """
    data = generate_synthetic_dataarray(
        n_sample, n_feature, index_policy, nan_policy, dask_policy, seed
    )
    dataset = xr.Dataset({"var0": data})
    seed += 1

    for n in range(1, n_variables):
        data_n = generate_synthetic_dataarray(
            n_sample=n_sample,
            n_feature=n_feature,
            index_policy=index_policy,
            nan_policy=nan_policy,
            dask_policy=dask_policy,
            seed=seed,
        )
        dataset[f"var{n}"] = data_n
        seed += 1
    return dataset


def generate_list_of_synthetic_dataarrays(
    n_arrays=1,
    n_sample=1,
    n_feature=1,
    index_policy="index",
    nan_policy="no_nan",
    dask_policy="no_dask",
    seed=0,
) -> DataList:
    """Create synthetic Dataset.

    Parameters:
    ------------
    n_arrays: int
        Number of DataArrays.
    n_sample: int
        Number of sample dimensions.
    n_dims_feature: int
        Number of feature dimensions.
    index_policy: ["index", "multiindex"], default="index"
        If "multiindex", the data will have a multiindex.
    nan_policy: ["no_nan", "isolated", "fulldim"], default="no_nan"
        If specified, the data will contain NaNs.
    dask_policy: ["no_dask", "dask"], default="no_dask"
        If "dask", the data will be a dask array.
    seed: int, default=0
        Seed for the random number generator.

    Returns:
    ---------
    data: xr.Dataset
        Synthetic data.

    """
    data_arrays = []
    for n in range(n_arrays):
        data_n = generate_synthetic_dataarray(
            n_sample=n_sample,
            n_feature=n_feature,
            index_policy=index_policy,
            nan_policy=nan_policy,
            dask_policy=dask_policy,
            seed=seed,
        )
        data_arrays.append(data_n)
        seed += 1
    return data_arrays


@pytest.fixture
def synthetic_dataarray(request) -> DataArray:
    data = generate_synthetic_dataarray(*request.param)
    return data


@pytest.fixture
def synthetic_dataset(request) -> DataSet:
    data = generate_synthetic_dataset(*request.param)
    return data


@pytest.fixture
def synthetic_datalist(request) -> DataList:
    data = generate_list_of_synthetic_dataarrays(*request.param)
    return data


# =============================================================================
# Input data
# =============================================================================
@pytest.fixture
def mock_data_array():
    rng = np.random.default_rng(7)
    noise = rng.normal(5, 3, size=(25, 5, 4))
    signal = 2 * np.sin(np.linspace(0, 2 * np.pi, 25))[:, None, None]
    return xr.DataArray(
        signal + noise,
        dims=("time", "lat", "lon"),
        coords={
            "time": xr.date_range("2001", "2025", freq="YS"),
            "lat": [20.0, 30.0, 40.0, 50.0, 60.0],
            "lon": [-10.0, 0.0, 10.0, 20.0],
        },
        name="t2m",
        attrs=dict(description="mock_data"),
    )


@pytest.fixture
def mock_dataset(mock_data_array):
    t2m = mock_data_array
    prcp = t2m**2
    return xr.Dataset({"t2m": t2m, "prcp": prcp})


@pytest.fixture
def mock_data_array_list(mock_data_array):
    da1 = mock_data_array
    da2 = mock_data_array**2
    da3 = mock_data_array**3
    return [da1, da2, da3]


@pytest.fixture
def mock_data_array_isolated_nans(mock_data_array):
    invalid_data = mock_data_array.copy()
    invalid_data.loc[dict(time="2001", lat=30.0, lon=-10.0)] = np.nan
    return invalid_data


@pytest.fixture
def mock_data_array_full_dimensional_nans(mock_data_array):
    valid_data = mock_data_array.copy()
    valid_data.loc[dict(lat=30.0)] = np.nan
    valid_data.loc[dict(time="2002")] = np.nan
    return valid_data


@pytest.fixture
def mock_data_array_boundary_nans(mock_data_array):
    valid_data = mock_data_array.copy(deep=True)
    valid_data.loc[dict(lat=30.0)] = np.nan
    valid_data.loc[dict(time="2001")] = np.nan
    return valid_data


@pytest.fixture
def mock_dask_data_array(mock_data_array):
    return mock_data_array.chunk({"lon": 2, "lat": 2, "time": -1})


@pytest.fixture
def mock_complex_data_array():
    def f1(x, t):
        return xr.DataArray(
            1.0 / np.cosh(x[np.newaxis, :] + 3) * np.exp(2.3j * t[:, np.newaxis]),
            coords=[("time", t), ("x", x)],
        )

    def f2(x, t):
        return xr.DataArray(
            2.0
            / np.cosh(x[np.newaxis, :])
            * np.tanh(x)
            * np.exp(2.8j * t[:, np.newaxis]),
            coords=[("time", t), ("x", x)],
        )

    x = np.linspace(-5, 5, 128)
    t = np.linspace(0, 4 * np.pi, 256)

    X1 = f1(x, t)
    X2 = f2(x, t)
    return X1 + X2


# =============================================================================
# Intermediate data
# =============================================================================
@pytest.fixture
def sample_input_data():
    """Create a sample input data."""
    return xr.DataArray(np.random.rand(10, 20), dims=("sample", "feature"))


@pytest.fixture
def sample_components():
    """Create a sample components."""
    return xr.DataArray(np.random.rand(20, 5), dims=("feature", "mode"))


@pytest.fixture
def sample_scores():
    """Create a sample scores."""
    return xr.DataArray(np.random.rand(10, 5), dims=("sample", "mode"))


@pytest.fixture
def sample_exp_var():
    return xr.DataArray(
        np.random.rand(10),
        dims=("mode",),
        coords={"mode": np.arange(10)},
        name="explained_variance",
    )


@pytest.fixture
def sample_total_variance(sample_exp_var):
    return sample_exp_var.sum()


@pytest.fixture
def sample_idx_modes_sorted(sample_exp_var):
    return sample_exp_var.argsort()[::-1]


@pytest.fixture
def sample_norm():
    return xr.DataArray(np.random.rand(10), dims=("mode",))


@pytest.fixture
def sample_squared_covariance():
    return xr.DataArray(np.random.rand(10), dims=("mode",))


@pytest.fixture
def sample_total_squared_covariance(sample_squared_covariance):
    return sample_squared_covariance.sum("mode")


@pytest.fixture
def sample_rotation_matrix():
    """Create a sample rotation matrix."""
    return xr.DataArray(np.random.rand(5, 5), dims=("mode_m", "mode_n"))


@pytest.fixture
def sample_phi_matrix():
    """Create a sample phi matrix."""
    return xr.DataArray(np.random.rand(5, 5), dims=("mode_m", "mode_n"))


@pytest.fixture
def sample_modes_sign():
    """Create a sample modes sign."""
    return xr.DataArray(np.random.choice([-1, 1], size=5), dims=("mode",))



from .cpcca_rotator import ComplexCPCCARotator, CPCCARotator, HilbertCPCCARotator
from .mca import MCA, ComplexMCA, HilbertMCA


class MCARotator(CPCCARotator, MCA):
    """Rotate a solution obtained from ``xe.cross.MCA``.

    Rotate the obtained components and scores of a CPCCA model to increase
    interpretability. The algorithm here is based on the approach of Cheng &
    Dunkerton (1995) [1]_.

    Parameters
    ----------
    n_modes : int, default=10
        Specify the number of modes to be rotated.
    power : int, default=1
        Set the power for the Promax rotation. A ``power`` value of 1 results in
        a Varimax rotation.
    max_iter : int or None, default=None
        Determine the maximum number of iterations for the computation of the
        rotation matrix. If not specified, defaults to 1000 if ``compute=True``
        and 100 if ``compute=False``, since we can't terminate a lazy
        computation based using ``rtol``.
    rtol : float, default=1e-8
        Define the relative tolerance required to achieve convergence and
        terminate the iterative process.
    compute : bool, default=True
        Whether to compute the rotation immediately.

    References
    ----------
    .. [1] Cheng, X. & Dunkerton, T. J. Orthogonal Rotation of Spatial Patterns
        Derived from Singular Value Decomposition Analysis. J. Climate 8,
        2631–2643 (1995).

    Examples
    --------

    Perform a MCA:

    >>> model = MCA(n_modes=10)
    >>> model.fit(X, Y, dim='time')

    Then, apply varimax rotation to first 5 components and scores:

    >>> rotator = MCARotator(n_modes=5)
    >>> rotator.fit(model)

    Retrieve the rotated components and scores:

    >>> rotator.components()
    >>> rotator.scores()

    """

    def __init__(
        self,
        n_modes: int = 10,
        power: int = 1,
        max_iter: int | None = None,
        rtol: float = 1e-8,
        compute: bool = True,
    ):
        CPCCARotator.__init__(
            self,
            n_modes=n_modes,
            power=power,
            max_iter=max_iter,
            rtol=rtol,
            compute=compute,
        )

        # Define analysis-relevant meta data
        self.attrs.update({"model": "Rotated MCA"})


class ComplexMCARotator(ComplexCPCCARotator, ComplexMCA):
    """Rotate a solution obtained from ``xe.cross.ComplexMCA``.

    Rotate the obtained components and scores of a CPCCA model to increase
    interpretability. The algorithm here is based on the approach of Cheng &
    Dunkerton (1995) [1]_, Elipot et al. (2017) [2]_ and Rieger et al. (2021).

    Parameters
    ----------
    n_modes : int, default=10
        Specify the number of modes to be rotated.
    power : int, default=1
        Set the power for the Promax rotation. A ``power`` value of 1 results in
        a Varimax rotation.
    max_iter : int, default=1000
        Determine the maximum number of iterations for the computation of the
        rotation matrix.
    rtol : float, default=1e-8
        Define the relative tolerance required to achieve convergence and
        terminate the iterative process.
    compute: bool, default=True
        Whether to compute the rotation immediately.

    References
    ----------
    .. [1] Cheng, X. & Dunkerton, T. J. Orthogonal Rotation of Spatial Patterns
        Derived from Singular Value Decomposition Analysis. J. Climate 8,
        2631–2643 (1995).
    .. [2] Elipot, S., Frajka-Williams, E., Hughes, C. W., Olhede, S. &
        Lankhorst, M. Observed Basin-Scale Response of the North Atlantic
        Meridional Overturning Circulation to Wind Stress Forcing. Journal of
        Climate 30, 2029–2054 (2017).
    .. [3] Rieger, N., Corral, Á., Olmedo, E. & Turiel, A. Lagged
        Teleconnections of Climate Variables Identified via Complex Rotated
        Maximum Covariance Analysis. Journal of Climate 34, 9861–9878 (2021).


    Examples
    --------

    Perform a Complex MCA:

    >>> model = ComplexMCA(n_modes=10)
    >>> model.fit(X, Y, dim='time')

    Then, apply varimax rotation to first 5 components and scores:

    >>> rotator = ComplexMCARotator(n_modes=5)
    >>> rotator.fit(model)

    Retrieve the rotated components and scores:

    >>> rotator.components()
    >>> rotator.scores()

    """

    def __init__(
        self,
        n_modes: int = 10,
        power: int = 1,
        max_iter: int | None = None,
        rtol: float = 1e-8,
        compute: bool = True,
    ):
        ComplexCPCCARotator.__init__(
            self,
            n_modes=n_modes,
            power=power,
            max_iter=max_iter,
            rtol=rtol,
            compute=compute,
        )
        self.attrs.update({"model": "Rotated Complex MCA"})


class HilbertMCARotator(HilbertCPCCARotator, HilbertMCA):
    """Rotate a solution obtained from ``xe.cross.HilbertMCA``.

    Rotate the obtained components and scores of a CPCCA model to increase
    interpretability. The algorithm here is based on the approach of Cheng &
    Dunkerton (1995) [1]_, Elipot et al. (2017) [2]_ and Rieger et al. (2021).

    Parameters
    ----------
    n_modes : int, default=10
        Specify the number of modes to be rotated.
    power : int, default=1
        Set the power for the Promax rotation. A ``power`` value of 1 results in
        a Varimax rotation.
    max_iter : int, default=1000
        Determine the maximum number of iterations for the computation of the
        rotation matrix.
    rtol : float, default=1e-8
        Define the relative tolerance required to achieve convergence and
        terminate the iterative process.
    compute: bool, default=True
        Whether to compute the rotation immediately.

    References
    ----------
    .. [1] Cheng, X. & Dunkerton, T. J. Orthogonal Rotation of Spatial Patterns
        Derived from Singular Value Decomposition Analysis. J. Climate 8,
        2631–2643 (1995).
    .. [2] Elipot, S., Frajka-Williams, E., Hughes, C. W., Olhede, S. &
        Lankhorst, M. Observed Basin-Scale Response of the North Atlantic
        Meridional Overturning Circulation to Wind Stress Forcing. Journal of
        Climate 30, 2029–2054 (2017).
    .. [3] Rieger, N., Corral, Á., Olmedo, E. & Turiel, A. Lagged
        Teleconnections of Climate Variables Identified via Complex Rotated
        Maximum Covariance Analysis. Journal of Climate 34, 9861–9878 (2021).


    Examples
    --------

    Perform a Hilbert MCA:

    >>> model = HilbertMCA(n_modes=10)
    >>> model.fit(X, Y, dim='time')

    Then, apply varimax rotation to first 5 components and scores:

    >>> rotator = HilbertMCARotator(n_modes=5)
    >>> rotator.fit(model)

    Retrieve the rotated components and scores:

    >>> rotator.components()
    >>> rotator.scores()

    """

    def __init__(
        self,
        n_modes: int = 10,
        power: int = 1,
        max_iter: int | None = None,
        rtol: float = 1e-8,
        compute: bool = True,
    ):
        HilbertCPCCARotator.__init__(
            self,
            n_modes=n_modes,
            power=power,
            max_iter=max_iter,
            rtol=rtol,
            compute=compute,
        )
        self.attrs.update({"model": "Rotated Hilbert MCA"})



from abc import abstractmethod
from typing import (
    Hashable,
    Sequence,
)

import xarray as xr
from typing_extensions import Self

from ..base_model import BaseModel
from ..data_container import DataContainer
from ..preprocessing.preprocessor import Preprocessor
from ..utils.data_types import DataArray, DataObject
from ..utils.sanity_checks import validate_input_type
from ..utils.xarray_utils import convert_to_dim_type

xr.set_options(keep_attrs=True)


class BaseModelSingleSet(BaseModel):
    """
    Abstract base class for single-set models.

    Parameters
    ----------
    n_modes: int, default=10
        Number of modes to calculate.
    center: bool, default=True
        Whether to center the input data.
    standardize: bool, default=False
        Whether to standardize the input data.
    use_coslat: bool, default=False
        Whether to use cosine of latitude for scaling.
    check_nans : bool, default=True
        If True, remove full-dimensional NaN features from the data, check to ensure
        that NaN features match the original fit data during transform, and check
        for isolated NaNs. Note: this forces eager computation of dask arrays.
        If False, skip all NaN checks. In this case, NaNs should be explicitly removed
        or filled prior to fitting, or SVD will fail.
    sample_name: str, default="sample"
        Name of the sample dimension.
    feature_name: str, default="feature"
        Name of the feature dimension.
    compute : bool, default=True
        Whether to compute elements of the model eagerly, or to defer computation.
        If True, four pieces of the fit will be computed sequentially: 1) the
        preprocessor scaler, 2) optional NaN checks, 3) SVD decomposition, 4) scores
        and components.
    random_state: int | None, default=None
        Seed for the random number generator.
    solver: {"auto", "full", "randomized"}, default="auto"
        Solver to use for the SVD computation.
    solver_kwargs: dict, default={}
        Additional keyword arguments to pass to the SVD solver function.

    """

    def __init__(
        self,
        n_modes=10,
        center=True,
        standardize=False,
        use_coslat=False,
        check_nans=True,
        sample_name="sample",
        feature_name="feature",
        compute=True,
        random_state=None,
        solver="auto",
        solver_kwargs={},
    ):
        super().__init__()

        self.n_modes = n_modes
        self.sample_name = sample_name
        self.feature_name = feature_name

        # Define model parameters
        self._params = {
            "n_modes": n_modes,
            "center": center,
            "standardize": standardize,
            "use_coslat": use_coslat,
            "check_nans": check_nans,
            "sample_name": sample_name,
            "feature_name": feature_name,
            "random_state": random_state,
            "compute": compute,
            "solver": solver,
            "solver_kwargs": solver_kwargs,
        }
        self._decomposer_kwargs = {
            "n_modes": n_modes,
            "solver": solver,
            "random_state": random_state,
            "compute": compute,
            "solver_kwargs": solver_kwargs,
        }

        # Define analysis-relevant meta data
        self.attrs.update({"model": "BaseModelSingleSet"})
        self.attrs.update(self._params)

        # Initialize the Preprocessor to scale and stack the data
        self.preprocessor = Preprocessor(
            sample_name=sample_name,
            feature_name=feature_name,
            with_center=center,
            with_std=standardize,
            with_coslat=use_coslat,
            check_nans=check_nans,
            compute=compute,
        )
        # Initialize the data container that stores the results
        self.data = DataContainer()

    def get_serialization_attrs(self) -> dict:
        return dict(
            data=self.data,
            preprocessor=self.preprocessor,
        )

    def fit(
        self,
        X: DataObject,
        dim: Sequence[Hashable] | Hashable,
        weights: DataObject | None = None,
    ) -> Self:
        """
        Fit the model to the input data.

        Parameters
        ----------
        X: DataObject
            Input data.
        dim: Sequence[Hashable] | Hashable
            Specify the sample dimensions. The remaining dimensions
            will be treated as feature dimensions.
        weights: DataObject | None, default=None
            Weighting factors for the input data.

        """
        # Check for invalid types
        validate_input_type(X)
        if weights is not None:
            validate_input_type(weights)

        self.sample_dims = convert_to_dim_type(dim)

        # Preprocess the data & transform to 2D
        data2D: DataArray = self.preprocessor.fit_transform(
            X, self.sample_dims, weights
        )

        self._fit_algorithm(data2D)

        if self._params["compute"]:
            self.data.compute()
            self._post_compute()

        return self

    @abstractmethod
    def _fit_algorithm(self, data: DataArray) -> Self:
        """Fit the model to the input data assuming a 2D DataArray.

        Parameters
        ----------
        data: DataArray
            Input data with dimensions (sample_name, feature_name)

        Returns
        -------
        self: Self
            The fitted model.

        """
        raise NotImplementedError

    def transform(self, data: DataObject, normalized=False) -> DataArray:
        """Project data onto the components.

        Parameters
        ----------
        data: DataObject
            Data to be transformed.
        normalized: bool, default=False
            Whether to normalize the scores by the L2 norm.

        Returns
        -------
        projections: DataArray
            Projections of the data onto the components.

        """
        validate_input_type(data)

        data2D = self.preprocessor.transform(data)
        data2D = self._transform_algorithm(data2D)
        if normalized:
            data2D = data2D / self.data["norms"]
            data2D.name = "scores"
        return self.preprocessor.inverse_transform_scores_unseen(data2D)

    @abstractmethod
    def _transform_algorithm(self, data: DataArray) -> DataArray:
        """Project data onto the components.

        Parameters
        ----------
        data: DataArray
            Input data with dimensions (sample_name, feature_name)

        Returns
        -------
        projections: DataArray
            Projections of the data onto the components.

        """
        raise NotImplementedError

    def fit_transform(
        self,
        data: DataObject,
        dim: Sequence[Hashable] | Hashable,
        weights: DataObject | None = None,
        **kwargs,
    ) -> DataArray:
        """Fit the model to the input data and project the data onto the components.

        Parameters
        ----------
        data: DataObject
            Input data.
        dim: Sequence[Hashable] | Hashable
            Specify the sample dimensions. The remaining dimensions
            will be treated as feature dimensions.
        weights: DataObject | None, default=None
            Weighting factors for the input data.
        **kwargs
            Additional keyword arguments to pass to the transform method.

        Returns
        -------
        projections: DataArray
            Projections of the data onto the components.

        """
        return self.fit(data, dim, weights).transform(data, **kwargs)

    def inverse_transform(
        self, scores: DataArray, normalized: bool = False
    ) -> DataObject:
        """Reconstruct the original data from transformed data.

        Parameters
        ----------
        scores: DataArray
            Transformed data to be reconstructed. This could be a subset
            of the `scores` data of a fitted model, or unseen data. Must
            have a 'mode' dimension.
        normalized: bool, default=False
            Whether the scores data have been normalized by the L2 norm.

        Returns
        -------
        data: DataObject
            Reconstructed data.

        """
        if normalized:
            norms = self.data["norms"].sel(mode=scores.mode)
            scores = scores * norms

        # Handle scalar mode in xr.dot
        if "mode" not in scores.dims:
            scores = scores.expand_dims("mode")

        data_reconstructed = self._inverse_transform_algorithm(scores)

        # Reconstructing the data using a single mode introduces a
        # redundant "mode" coordinate
        if "mode" in data_reconstructed.coords:
            data_reconstructed = data_reconstructed.drop_vars("mode")

        return self.preprocessor.inverse_transform_data(data_reconstructed)

    @abstractmethod
    def _inverse_transform_algorithm(self, scores: DataArray) -> DataArray:
        """Reconstruct the original data from transformed data.

        Parameters
        ----------
        scores: DataObject
            Transformed data to be reconstructed. This could be a subset
            of the `scores` data of a fitted model, or unseen data. Must
            have a 'mode' dimension.

        Returns
        -------
        data: DataArray
            Reconstructed 2D data with dimensions (sample_name, feature_name)

        """
        raise NotImplementedError

    def components(self, normalized: bool = True) -> DataObject:
        """Get the components.

        Parameters
        ----------
        normalized: bool, default=True
            Whether to normalize the components by the L2 norm.

        """
        components = self.data["components"]
        if not normalized:
            name = components.name
            components = components * self.data["norms"]
            components.name = name
        return self.preprocessor.inverse_transform_components(components)

    def scores(self, normalized: bool = False) -> DataArray:
        """Get the scores.

        Parameters
        ----------
        normalized: bool, default=True
            Whether to normalize the scores by the L2 norm.
        """
        scores = self.data["scores"].copy()
        if normalized:
            name = scores.name
            scores = scores / self.data["norms"]
            scores.name = name
        return self.preprocessor.inverse_transform_scores(scores)



import dask.array as da
import numpy as np
import pytest
import xarray as xr

from xeofs.cross import CPCCA

from ...utilities import skip_if_missing_engine


def generate_random_data(shape, lazy=False, seed=142):
    rng = np.random.default_rng(seed)
    if lazy:
        return xr.DataArray(
            da.random.random(shape, chunks=(5, 5)),
            dims=["sample", "feature"],
            coords={"sample": np.arange(shape[0]), "feature": np.arange(shape[1])},
        )
    else:
        return xr.DataArray(
            rng.random(shape),
            dims=["sample", "feature"],
            coords={"sample": np.arange(shape[0]), "feature": np.arange(shape[1])},
        )


def generate_well_conditioned_data(lazy=False):
    rng = np.random.default_rng(142)
    t = np.linspace(0, 50, 200)
    std = 0.1
    x1 = np.sin(t)[:, None] + rng.normal(0, std, size=(200, 2))
    x2 = np.sin(t)[:, None] + rng.normal(0, std, size=(200, 3))
    x1[:, 1] = x1[:, 1] ** 2
    x2[:, 1] = x2[:, 1] ** 3
    x2[:, 2] = abs(x2[:, 2]) ** (0.5)
    coords_time = np.arange(len(t))
    coords_fx = [1, 2]
    coords_fy = [1, 2, 3]
    X = xr.DataArray(
        x1,
        dims=["sample", "feature"],
        coords={"sample": coords_time, "feature": coords_fx},
    )
    Y = xr.DataArray(
        x2,
        dims=["sample", "feature"],
        coords={"sample": coords_time, "feature": coords_fy},
    )
    if lazy:
        X = X.chunk({"sample": 5, "feature": -1})
        Y = Y.chunk({"sample": 5, "feature": -1})
        return X, Y
    else:
        return X, Y


@pytest.fixture
def cpcca():
    return CPCCA(n_modes=1)


def test_initialization():
    model = CPCCA()
    assert model is not None


def test_fit(cpcca):
    X, Y = generate_well_conditioned_data()
    cpcca.fit(X, Y, dim="sample")
    assert hasattr(cpcca, "preprocessor1")
    assert hasattr(cpcca, "preprocessor2")
    assert hasattr(cpcca, "data")


def test_fit_empty_data(cpcca):
    with pytest.raises(ValueError):
        cpcca.fit(xr.DataArray(), xr.DataArray(), "time")


def test_fit_invalid_dims(cpcca):
    X, Y = generate_well_conditioned_data()
    with pytest.raises(ValueError):
        cpcca.fit(X, Y, dim=("invalid_dim1", "invalid_dim2"))


def test_transform(cpcca):
    X, Y = generate_well_conditioned_data()
    cpcca.fit(X, Y, dim="sample")
    result = cpcca.transform(X, Y)
    assert isinstance(result, list)
    assert isinstance(result[0], xr.DataArray)


def test_transform_unseen_data(cpcca):
    X, Y = generate_well_conditioned_data()
    x = X.isel(sample=slice(151, 200))
    y = Y.isel(sample=slice(151, 200))
    X = X.isel(sample=slice(None, 150))
    Y = Y.isel(sample=slice(None, 150))

    cpcca.fit(X, Y, "sample")
    result = cpcca.transform(x, y)
    assert isinstance(result, list)
    assert isinstance(result[0], xr.DataArray)
    # Check that unseen data can be transformed
    assert result[0].notnull().all()
    assert result[1].notnull().all()


def test_inverse_transform(cpcca):
    X, Y = generate_well_conditioned_data()
    cpcca.fit(X, Y, "sample")
    # Assuming mode as 1 for simplicity
    scores1 = cpcca.data["scores1"].sel(mode=1)
    scores2 = cpcca.data["scores2"].sel(mode=1)
    Xrec1, Xrec2 = cpcca.inverse_transform(scores1, scores2)
    assert isinstance(Xrec1, xr.DataArray)
    assert isinstance(Xrec2, xr.DataArray)


@pytest.mark.parametrize(
    "alpha,use_pca",
    [
        (1.0, False),
        (0.5, False),
        (0.0, False),
        (1.0, True),
        (0.5, True),
        (0.0, True),
    ],
)
def test_squared_covariance_fraction(alpha, use_pca):
    X, Y = generate_well_conditioned_data()
    cpcca = CPCCA(n_modes=2, alpha=alpha, use_pca=use_pca, n_pca_modes="all")
    cpcca.fit(X, Y, "sample")
    scf = cpcca.squared_covariance_fraction()
    assert isinstance(scf, xr.DataArray)
    assert all(scf <= 1), "Squared covariance fraction is greater than 1"


@pytest.mark.parametrize(
    "alpha,use_pca",
    [
        (1.0, False),
        (0.5, False),
        (0.0, False),
        (1.0, True),
        (0.5, True),
        (0.0, True),
    ],
)
def test_total_squared_covariance(alpha, use_pca):
    X, Y = generate_well_conditioned_data()

    # Compute total squared covariance
    X_ = X.rename({"feature": "x"})
    Y_ = Y.rename({"feature": "y"})
    cov_mat = xr.cov(X_, Y_, dim="sample")
    tsc = (cov_mat**2).sum()

    cpcca = CPCCA(n_modes=2, alpha=alpha, use_pca=use_pca, n_pca_modes="all")
    cpcca.fit(X, Y, "sample")
    tsc_model = cpcca.data["total_squared_covariance"]
    xr.testing.assert_allclose(tsc, tsc_model)


def test_alpha_integer():
    X, Y = generate_well_conditioned_data()

    cpcca = CPCCA(n_modes=2, alpha=1, use_pca=False)
    cpcca.fit(X, Y, "sample")


def test_fit_different_coordinates():
    """Like a lagged CCA scenario"""
    X, Y = generate_well_conditioned_data()
    X = X.isel(sample=slice(0, 99))
    Y = Y.isel(sample=slice(100, 199))
    cpcca = CPCCA(n_modes=2, alpha=1, use_pca=False)
    cpcca.fit(X, Y, "sample")
    r = cpcca.cross_correlation_coefficients()
    # Correlation coefficents are not zero
    assert np.all(r > np.finfo(r.dtype).eps)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(mock_data_array, dim):
    cpcca = CPCCA(n_modes=2, alpha=1, use_pca=False)
    cpcca.fit(mock_data_array, mock_data_array, dim)
    components1, components2 = cpcca.components()
    feature_dims = tuple(set(mock_data_array.dims) - set(dim))
    assert isinstance(components1, xr.DataArray)
    assert isinstance(components2, xr.DataArray)
    assert set(components1.dims) == set(
        ("mode",) + feature_dims
    ), "Components1 does not have the right feature dimensions"
    assert set(components2.dims) == set(
        ("mode",) + feature_dims
    ), "Components2 does not have the right feature dimensions"


@pytest.mark.parametrize(
    "shapeX,shapeY,alpha,use_pca",
    [
        ((20, 10), (20, 10), 1.0, False),
        ((20, 40), (20, 30), 1.0, False),
        ((20, 10), (20, 40), 1.0, False),
        ((20, 10), (20, 10), 0.5, False),
        ((20, 40), (20, 30), 0.5, False),
        ((20, 10), (20, 40), 0.5, False),
        ((20, 10), (20, 10), 1.0, True),
        ((20, 40), (20, 30), 1.0, True),
        ((20, 10), (20, 40), 1.0, True),
        ((20, 10), (20, 10), 0.5, True),
        ((20, 40), (20, 30), 0.5, True),
        ((20, 10), (20, 40), 0.5, True),
    ],
)
def test_components_coordinates(shapeX, shapeY, alpha, use_pca):
    # Test that the components have the right coordinates
    X = generate_random_data(shapeX)
    Y = generate_random_data(shapeY)

    cpcca = CPCCA(n_modes=2, alpha=alpha, use_pca=use_pca, n_pca_modes="all")
    cpcca.fit(X, Y, "sample")
    components1, components2 = cpcca.components()
    xr.testing.assert_equal(components1.coords["feature"], X.coords["feature"])
    xr.testing.assert_equal(components2.coords["feature"], Y.coords["feature"])


@pytest.mark.parametrize(
    "correction",
    [(None), ("fdr_bh")],
)
def test_homogeneous_patterns(correction):
    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    cpcca = CPCCA(n_modes=10, alpha=1, use_pca=False)
    cpcca.fit(X, Y, "sample")

    _ = cpcca.homogeneous_patterns(correction=correction)


@pytest.mark.parametrize(
    "correction",
    [(None), ("fdr_bh")],
)
def test_heterogeneous_patterns(correction):
    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    cpcca = CPCCA(n_modes=10, alpha=1, use_pca=False)
    cpcca.fit(X, Y, "sample")

    _ = cpcca.heterogeneous_patterns(correction=correction)


def test_predict():
    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    cpcca = CPCCA(n_modes=10, alpha=0.2, use_pca=False)
    cpcca.fit(X, Y, "sample")

    Xnew = generate_random_data((200, 10), seed=123)

    Ry_pred = cpcca.predict(Xnew)
    _ = cpcca.inverse_transform(Y=Ry_pred)


@pytest.mark.parametrize("engine", ["h5netcdf", "netcdf4", "zarr"])
@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.0])
def test_save_load(tmp_path, engine, alpha):
    """Test save/load methods in MCA class, ensuring that we can
    roundtrip the model and get the same results when transforming
    data."""
    skip_if_missing_engine(engine)

    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    original = CPCCA(alpha=alpha)
    original.fit(X, Y, "sample")

    # Save the CPCCA model
    original.save(tmp_path / "cpcca", engine=engine)

    # Check that the CPCCA model has been saved
    assert (tmp_path / "cpcca").exists()

    # Recreate the model from saved file
    loaded = CPCCA.load(tmp_path / "cpcca", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        if original.data._allow_compute[key]:
            assert loaded.data[key].equals(original.data[key])
        else:
            # but ensure that input data is not saved by default
            assert loaded.data[key].size <= 1
            assert loaded.data[key].attrs["placeholder"] is True

    # Test that the recreated model can be used to transform new data
    assert np.allclose(
        original.transform(X, Y),
        loaded.transform(X, Y),
    )

    # The loaded model should also be able to inverse_transform new data
    XYr_o = original.inverse_transform(*original.scores())
    XYr_l = loaded.inverse_transform(*loaded.scores())
    assert np.allclose(XYr_o[0], XYr_l[0])
    assert np.allclose(XYr_o[1], XYr_l[1])


@pytest.mark.parametrize("engine", ["h5netcdf", "netcdf4", "zarr"])
@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.0])
def test_save_load_with_data(tmp_path, engine, alpha):
    """Test save/load methods in CPCCA class, ensuring that we can
    roundtrip the model and get the same results for SCF."""
    skip_if_missing_engine(engine)

    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    original = CPCCA(alpha=alpha)
    original.fit(X, Y, "sample")

    # Save the CPCCA model
    original.save(tmp_path / "cpcca", engine=engine, save_data=True)

    # Check that the CPCCA model has been saved
    assert (tmp_path / "cpcca").exists()

    # Recreate the model from saved file
    loaded = CPCCA.load(tmp_path / "cpcca", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        assert loaded.data[key].equals(original.data[key])

    # Test that the recreated model can compute the SCF
    assert np.allclose(
        original.squared_covariance_fraction(), loaded.squared_covariance_fraction()
    )


def test_serialize_deserialize_dataarray(mock_data_array):
    """Test roundtrip serialization when the model is fit on a DataArray."""
    model = CPCCA()
    model.fit(mock_data_array, mock_data_array, "time")
    dt = model.serialize()
    rebuilt_model = CPCCA.deserialize(dt)
    assert np.allclose(
        model.transform(mock_data_array), rebuilt_model.transform(mock_data_array)
    )


def test_serialize_deserialize_dataset(mock_dataset):
    """Test roundtrip serialization when the model is fit on a Dataset."""
    model = CPCCA()
    model.fit(mock_dataset, mock_dataset, "time")
    dt = model.serialize()
    rebuilt_model = CPCCA.deserialize(dt)
    assert np.allclose(
        model.transform(mock_dataset), rebuilt_model.transform(mock_dataset)
    )



import pytest

from xeofs.cross import MCA, HilbertMCA, HilbertMCARotator, MCARotator
from xeofs.rotator_factory import RotatorFactory
from xeofs.single import EOF, EOFRotator, HilbertEOF, HilbertEOFRotator

# RotatorFactory should be imported from its module
# from module import RotatorFactory


def test_rotator_factory_init():
    factory = RotatorFactory(n_modes=3, power=2, max_iter=1000, rtol=1e-8)
    assert factory.params == {"n_modes": 3, "power": 2, "max_iter": 1000, "rtol": 1e-8}
    assert factory._valid_types == (EOF, HilbertEOF, MCA, HilbertMCA)


def test_create_rotator_EOF():
    factory = RotatorFactory(n_modes=3, power=2, max_iter=1000, rtol=1e-8)
    EOF_instance = EOF()
    rotator = factory.create_rotator(EOF_instance)
    assert isinstance(rotator, EOFRotator)


def test_create_rotator_HilbertEOF():
    factory = RotatorFactory(n_modes=3, power=2, max_iter=1000, rtol=1e-8)
    HilbertEOF_instance = HilbertEOF()  # creating instance of the mock class
    rotator = factory.create_rotator(HilbertEOF_instance)
    assert isinstance(rotator, HilbertEOFRotator)


def test_create_rotator_MCA():
    pytest.importorskip("statsmodels")
    factory = RotatorFactory(n_modes=3, power=2, max_iter=1000, rtol=1e-8)
    MCA_instance = MCA()
    rotator = factory.create_rotator(MCA_instance)
    assert isinstance(rotator, MCARotator)


def test_create_rotator_HilbertMCA():
    pytest.importorskip("statsmodels")
    factory = RotatorFactory(n_modes=3, power=2, max_iter=1000, rtol=1e-8)
    HilbertMCA_instance = HilbertMCA()
    rotator = factory.create_rotator(HilbertMCA_instance)
    assert isinstance(rotator, HilbertMCARotator)


def test_create_rotator_invalid_model():
    factory = RotatorFactory(n_modes=3, power=2, max_iter=1000, rtol=1e-8)
    with pytest.raises(TypeError):
        factory.create_rotator("InvalidType")  # type: ignore



import numpy as np
import xarray as xr

from xeofs.utils.xarray_utils import total_variance


def test_total_variance(mock_data_array):
    """Test the total_variance function."""
    arr = mock_data_array.copy().values
    # Compute total variance using numpy by hand
    tot_var_ref = np.sum(np.var(arr, axis=0, ddof=1))

    da = mock_data_array.stack(sample=("time",), feature=("lat", "lon"))
    da -= da.mean(dim="sample")
    tot_var = total_variance(da, dim="sample")
    assert isinstance(tot_var, xr.DataArray)
    assert tot_var.dims == ()
    assert tot_var.shape == ()
    assert tot_var.dtype == np.float64
    assert np.allclose(
        tot_var, tot_var_ref
    ), "Total variance computed by hand does not match."



import pytest
import numpy as np

from xeofs.preprocessing.preprocessor import Preprocessor
from ..conftest import generate_synthetic_dataarray
from ..utilities import (
    get_dims_from_data,
    data_is_dask,
    data_has_multiindex,
    assert_expected_dims,
    assert_expected_coords,
)

# =============================================================================
# GENERALLY VALID TEST CASES
# =============================================================================
N_SAMPLE_DIMS = [1, 2]
N_FEATURE_DIMS = [1, 2]
INDEX_POLICY = ["index", "multiindex"]
NAN_POLICY = ["no_nan", "fulldim"]
DASK_POLICY = ["no_dask", "dask"]
SEED = [0]

TEST_DATA_PARAMS = [
    (ns, nf, index, nan, dask)
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in NAN_POLICY
    for dask in DASK_POLICY
]

SAMPLE_DIM_NAMES = ["sample", "sample_alternative"]
FEATURE_DIM_NAMES = ["feature", "feature_alternative"]

VALID_TEST_CASES = [
    (sample_name, feature_name, data_params)
    for sample_name in SAMPLE_DIM_NAMES
    for feature_name in FEATURE_DIM_NAMES
    for data_params in TEST_DATA_PARAMS
]


# TESTS
# =============================================================================
@pytest.mark.parametrize(
    "with_std, with_coslat, with_weights",
    [
        (True, True, True),
        (True, True, False),
        (True, False, True),
        (True, False, False),
        (False, True, True),
        (False, True, False),
        (False, False, True),
        (False, False, False),
    ],
)
def test_fit_transform_scalings(with_std, with_coslat, with_weights, mock_data_array):
    """fit method should not be implemented."""
    prep = Preprocessor(with_std=with_std, with_coslat=with_coslat)

    weights = None
    if with_weights:
        weights = mock_data_array.mean("time").copy()
        weights[:] = 0.5

    data_trans = prep.fit_transform(
        mock_data_array,
        weights=weights,
        sample_dims=("time",),
    )

    assert hasattr(prep, "scaler")
    assert hasattr(prep, "preconverter")
    assert hasattr(prep, "stacker")
    assert hasattr(prep, "postconverter")
    assert hasattr(prep, "sanitizer")

    # Transformed data is centered
    assert np.isclose(data_trans.mean("sample"), 0).all()
    # Transformed data is standardized
    if with_std and not with_coslat:
        if with_weights:
            assert np.isclose(data_trans.std("sample"), 0.5).all()
        else:
            assert np.isclose(data_trans.std("sample"), 1).all()


@pytest.mark.parametrize(
    "index_policy, nan_policy, dask_policy",
    [
        ("index", "no_nan", "no_dask"),
        ("multiindex", "no_nan", "dask"),
        ("index", "fulldim", "no_dask"),
        ("multiindex", "fulldim", "dask"),
    ],
)
def test_fit_transform_same_dim_names(index_policy, nan_policy, dask_policy):
    data = generate_synthetic_dataarray(1, 1, index_policy, nan_policy, dask_policy)
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    prep = Preprocessor(sample_name="sample0", feature_name="feature0")
    transformed = prep.fit_transform(data, sample_dims)
    reconstructed = prep.inverse_transform_data(transformed)

    data_is_dask_before = data_is_dask(data)
    data_is_dask_interm = data_is_dask(transformed)
    data_is_dask_after = data_is_dask(reconstructed)

    assert set(transformed.dims) == set(("sample0", "feature0"))
    assert set(reconstructed.dims) == set(("sample0", "feature0"))
    assert not data_has_multiindex(transformed)
    assert transformed.notnull().all()
    assert data_is_dask_before == data_is_dask_interm
    assert data_is_dask_before == data_is_dask_after


@pytest.mark.parametrize(
    "sample_name, feature_name, data_params",
    VALID_TEST_CASES,
)
def test_fit_transform(sample_name, feature_name, data_params):
    data = generate_synthetic_dataarray(*data_params)
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    prep = Preprocessor(sample_name=sample_name, feature_name=feature_name)
    transformed = prep.fit_transform(data, sample_dims)

    data_is_dask_before = data_is_dask(data)
    data_is_dask_after = data_is_dask(transformed)

    assert transformed.dims == (sample_name, feature_name)
    assert not data_has_multiindex(transformed)
    assert transformed.notnull().all()
    assert data_is_dask_before == data_is_dask_after


@pytest.mark.parametrize(
    "sample_name, feature_name, data_params",
    VALID_TEST_CASES,
)
def test_inverse_transform(sample_name, feature_name, data_params):
    data = generate_synthetic_dataarray(*data_params)
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    prep = Preprocessor(sample_name=sample_name, feature_name=feature_name)
    transformed = prep.fit_transform(data, sample_dims)
    components = transformed.rename({sample_name: "mode"})
    scores = transformed.rename({feature_name: "mode"})

    reconstructed = prep.inverse_transform_data(transformed)
    components = prep.inverse_transform_components(components)
    scores = prep.inverse_transform_scores(scores)

    # Reconstructed data has the same dimensions as the original data
    assert_expected_dims(data, reconstructed, policy="all")
    assert_expected_dims(data, components, policy="feature")
    assert_expected_dims(data, scores, policy="sample")

    # Reconstructed data has the same coordinates as the original data
    assert_expected_coords(data, reconstructed, policy="all")
    assert_expected_coords(data, components, policy="feature")
    assert_expected_coords(data, scores, policy="sample")

    # Reconstructed data and original data have NaNs in the same FEATURES
    # Note: NaNs in the same place is not guaranteed, since isolated NaNs will be propagated
    # to all samples in the same feature
    features_with_nans_before = data.isnull().any(sample_dims)
    features_with_nans_after = reconstructed.isnull().any(sample_dims)
    assert features_with_nans_before.equals(features_with_nans_after)

    # Reconstructed data has MultiIndex if and only if original data has MultiIndex
    data_has_multiindex_before = data_has_multiindex(data)
    data_has_multiindex_after = data_has_multiindex(reconstructed)
    assert data_has_multiindex_before == data_has_multiindex_after

    # Reconstructed data is dask if and only if original data is dask
    data_is_dask_before = data_is_dask(data)
    data_is_dask_after = data_is_dask(reconstructed)
    assert data_is_dask_before == data_is_dask_after



import pytest

from xeofs.multi import CCA


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(dim, mock_data_array_list):
    """Tests the fit method of the CCA class"""

    cca = CCA()
    cca.fit(mock_data_array_list, dim)

    # Assert the required attributes have been set
    assert hasattr(cca, "preprocessors")
    assert hasattr(cca, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(dim, mock_data_array_list):
    """Tests the components method of the CCA class"""

    cca = CCA()
    cca.fit(mock_data_array_list, dim)

    comps = cca.components()
    assert isinstance(comps, list)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores(dim, mock_data_array_list):
    """Tests the components method of the CCA class"""

    cca = CCA()
    cca.fit(mock_data_array_list, dim)

    scores = cca.scores()
    assert isinstance(scores, list)



import warnings

from .cca import CCA, ComplexCCA, HilbertCCA
from .cpcca import CPCCA, ComplexCPCCA, HilbertCPCCA
from .cpcca_rotator import ComplexCPCCARotator, CPCCARotator, HilbertCPCCARotator
from .mca import MCA, ComplexMCA, HilbertMCA
from .mca_rotator import ComplexMCARotator, HilbertMCARotator, MCARotator
from .rda import RDA, ComplexRDA, HilbertRDA

__all__ = [
    "CCA",
    "MCA",
    "RDA",
    "CPCCA",
    "ComplexCCA",
    "ComplexMCA",
    "ComplexRDA",
    "ComplexCPCCA",
    "HilbertCCA",
    "HilbertMCA",
    "HilbertRDA",
    "HilbertCPCCA",
    "MCARotator",
    "CPCCARotator",
    "ComplexMCARotator",
    "ComplexCPCCARotator",
    "HilbertMCARotator",
    "HilbertCPCCARotator",
]


DEPRECATED_NAMES = [
    # ("OldClass", "NewClass"),
]


def __dir__():
    return sorted(__all__ + [names[0] for names in DEPRECATED_NAMES])


def __getattr__(name):
    for old_name, new_name in DEPRECATED_NAMES:
        if name == old_name:
            msg = (
                f"Class '{old_name}' is deprecated and will be renamed to '{new_name}' in the next major release. "
                f"In that release, '{old_name}' will refer to a different class. "
                f"Please switch to '{new_name}' to maintain compatibility."
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return globals()[new_name]
    raise AttributeError(f"module {__name__} has no attribute {name}")



import pandas as pd
from typing_extensions import Self

from ..utils.data_types import Data, DataArray, DataVar, DataVarBound, Dims
from .transformer import Transformer


class MultiIndexConverter(Transformer):
    """Convert MultiIndexes of an ND DataArray or Dataset to regular indexes."""

    def __init__(self):
        super().__init__()
        self.modified_dimensions = []
        self.coords_from_fit = {}
        self.coords_from_transform = {}

    def get_serialization_attrs(self) -> dict:
        return dict(
            modified_dimensions=self.modified_dimensions,
            coords_from_fit=self.coords_from_fit,
            coords_from_transform=self.coords_from_transform,
        )

    def fit(
        self,
        X: Data,
        sample_dims: Dims | None = None,
        feature_dims: Dims | None = None,
        **kwargs,
    ) -> Self:
        # Store original MultiIndexes
        for dim in X.dims:
            index = X.indexes[dim]
            if isinstance(index, pd.MultiIndex):
                self.coords_from_fit[dim] = X.coords[dim]
                self.modified_dimensions.append(dim)

        return self

    def transform(self, X: DataVar) -> DataVar:
        X_transformed = X.copy(deep=True)

        # Replace MultiIndexes with simple index
        for dim in self.modified_dimensions:
            # We need to store the indexes from "unseen" data
            self.coords_from_transform[dim] = X_transformed.coords[dim]

            index = X_transformed.indexes[dim]
            X_transformed = X_transformed.drop_vars(dim)
            X_transformed.coords[dim] = range(index.size)

        return X_transformed

    def _inverse_transform(self, X: DataVarBound, reference: str) -> DataVarBound:
        X_inverse_transformed = X.copy(deep=True)

        match reference:
            case "fit":
                reference_indexes = self.coords_from_fit
            case "transform":
                reference_indexes = self.coords_from_transform

        # Restore original MultiIndexes
        for dim, original_index in reference_indexes.items():
            if dim in X_inverse_transformed.dims:
                X_inverse_transformed.coords[dim] = original_index
                # Set indexes to original MultiIndexes
                indexes = [idx for idx in original_index.indexes.keys() if idx != dim]
                X_inverse_transformed = X_inverse_transformed.set_index({dim: indexes})

        return X_inverse_transformed

    def inverse_transform_data(self, X: DataVarBound) -> DataVarBound:
        return self._inverse_transform(X, reference="fit")

    def inverse_transform_components(self, X: DataVarBound) -> DataVarBound:
        return self._inverse_transform(X, reference="fit")

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        return self._inverse_transform(X, reference="fit")

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        return self._inverse_transform(X, reference="transform")



import dask.array as da
import numpy as np
import pytest
import xarray as xr

from xeofs.cross import HilbertCPCCA

from ...utilities import skip_if_missing_engine


def generate_random_data(shape, lazy=False, seed=142):
    rng = np.random.default_rng(seed)
    if lazy:
        return xr.DataArray(
            da.random.random(shape, chunks=(5, 5)),
            dims=["sample", "feature"],
            coords={"sample": np.arange(shape[0]), "feature": np.arange(shape[1])},
        )
    else:
        return xr.DataArray(
            rng.random(shape),
            dims=["sample", "feature"],
            coords={"sample": np.arange(shape[0]), "feature": np.arange(shape[1])},
        )


def generate_well_conditioned_data(lazy=False):
    rng = np.random.default_rng(142)
    t = np.linspace(0, 50, 200)
    std = 0.1
    x1 = np.sin(t)[:, None] + rng.normal(0, std, size=(200, 2))
    x2 = np.sin(t)[:, None] + rng.normal(0, std, size=(200, 3))
    x1[:, 1] = x1[:, 1] ** 2
    x2[:, 1] = x2[:, 1] ** 3
    x2[:, 2] = abs(x2[:, 2]) ** (0.5)
    coords_time = np.arange(len(t))
    coords_fx = [1, 2]
    coords_fy = [1, 2, 3]
    X = xr.DataArray(
        x1,
        dims=["sample", "feature"],
        coords={"sample": coords_time, "feature": coords_fx},
    )
    Y = xr.DataArray(
        x2,
        dims=["sample", "feature"],
        coords={"sample": coords_time, "feature": coords_fy},
    )
    if lazy:
        X = X.chunk({"sample": 5, "feature": -1})
        Y = Y.chunk({"sample": 5, "feature": -1})
        return X, Y
    else:
        return X, Y


@pytest.mark.parametrize("use_pca", [True, False])
def test_singular_values(use_pca):
    """Test that the singular values of the Hilbert CCA are less than 1."""
    X, Y = generate_well_conditioned_data()
    cpcca = HilbertCPCCA(n_modes=2, alpha=0.0, use_pca=use_pca, n_pca_modes=2)
    cpcca.fit(X, Y, "sample")
    s_values = cpcca.data["singular_values"]

    # Singular values are the canonical correlations, so they should be less than 1
    assert np.all(s_values <= 1)


# Currently, netCDF4 does not support complex numbers, so skip this test
@pytest.mark.parametrize("engine", ["h5netcdf", "zarr"])
@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.0])
def test_save_load_with_data(tmp_path, engine, alpha):
    """Test save/load methods in CPCCA class, ensuring that we can
    roundtrip the model and get the same results."""
    skip_if_missing_engine(engine)

    X = generate_random_data((200, 10), seed=123)
    Y = generate_random_data((200, 20), seed=321)

    original = HilbertCPCCA(alpha=alpha)
    original.fit(X, Y, "sample")

    # Save the CPCCA model
    original.save(tmp_path / "cpcca", engine=engine, save_data=True)

    # Check that the CPCCA model has been saved
    assert (tmp_path / "cpcca").exists()

    # Recreate the model from saved file
    loaded = HilbertCPCCA.load(tmp_path / "cpcca", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        assert loaded.data[key].equals(original.data[key])

    # Test that the recreated model can compute the SCF
    assert np.allclose(
        original.squared_covariance_fraction(), loaded.squared_covariance_fraction()
    )

    # Test that the recreated model can compute the components amplitude
    A1_original, A2_original = original.components_amplitude()
    A1_loaded, A2_loaded = loaded.components_amplitude()
    assert np.allclose(A1_original, A1_loaded)
    assert np.allclose(A2_original, A2_loaded)

    # Test that the recreated model can compute the components phase
    P1_original, P2_original = original.components_phase()
    P1_loaded, P2_loaded = loaded.components_phase()
    assert np.allclose(P1_original, P1_loaded)
    assert np.allclose(P2_original, P2_loaded)



import numpy as np
import xarray as xr
from typing_extensions import Self

from ..utils.data_types import (
    Data,
    DataArray,
    Dims,
    DimsList,
)
from ..utils.xarray_utils import (
    _check_parameter_number,
    convert_to_list,
    get_dims,
    process_parameter,
    unwrap_singleton_list,
)
from .concatenator import Concatenator
from .dimension_renamer import DimensionRenamer
from .list_processor import GenericListTransformer
from .multi_index_converter import MultiIndexConverter
from .sanitizer import Sanitizer
from .scaler import Scaler
from .stacker import Stacker
from .transformer import Transformer


def extract_new_dim_names(X: list[DimensionRenamer]) -> tuple[Dims, DimsList]:
    """Extract the new dimension names from a list of DimensionRenamer objects.

    Parameters
    ----------
    X : list of DimensionRenamer
        list of DimensionRenamer objects.

    Returns
    -------
    Dims
        Sample dimensions
    DimsList
        Feature dimenions

    """
    new_sample_dims = []
    new_feature_dims: DimsList = []
    for x in X:
        new_sample_dims.append(x.sample_dims_after)
        new_feature_dims.append(x.feature_dims_after)
    new_sample_dims: Dims = tuple(np.unique(np.asarray(new_sample_dims)).tolist())
    return new_sample_dims, new_feature_dims


class Preprocessor(Transformer):
    """Preprocess xarray objects (DataArray, Dataset).

    Preprocessing includes
        (i) Feature-wise scaling (e.g. removing mean, dividing by standard deviation, applying (latitude) weights
        (ii) Renaming dimensions (to avoid conflicts with sample and feature dimensions)
        (iii) Converting MultiIndexes to regular Indexes (MultiIndexes cannot be stacked)
        (iv) Stacking the data into 2D DataArray
        (v) Converting MultiIndexes introduced by stacking into regular Indexes
        (vi) Removing NaNs
        (vii) Concatenating the 2D DataArrays into one 2D DataArray

    Parameters
    ----------
    sample_name : str, default="sample"
        Name of the sample dimension.
    feature_name : str, default="feature"
        Name of the feature dimension.
    with_center : bool, default=True
        If True, the data is centered by subtracting the mean.
    with_std : bool, default=True
        If True, the data is divided by the standard deviation.
    with_coslat : bool, default=False
        If True, the data is multiplied by the square root of cosine of latitude weights.
    with_weights : bool, default=False
        If True, the data is multiplied by additional user-defined weights.
    return_list : bool, default=True
        If True, inverse_transform methods returns always a list of DataArray(s).
        If False, the output is returned as a single DataArray if possible.
    check_nans : bool, default=True
        If True, remove full-dimensional NaN features from the data, check to ensure
        that NaN features match the original fit data during transform, and check
        for isolated NaNs. Note: this forces eager computation of dask arrays.
        If False, skip all NaN checks. In this case, NaNs should be explicitly removed
        or filled prior to fitting, or SVD will fail.

    """

    def __init__(
        self,
        sample_name: str = "sample",
        feature_name: str = "feature",
        with_center: bool = True,
        with_std: bool = False,
        with_coslat: bool = False,
        return_list: bool = True,
        check_nans: bool = True,
        compute: bool = True,
    ):
        # Set parameters
        self.sample_name = sample_name
        self.feature_name = feature_name
        self.with_center = with_center
        self.with_std = with_std
        self.with_coslat = with_coslat
        self.return_list = return_list
        self.check_nans = check_nans
        self.compute = compute

        self.n_data = None

        dim_names_as_kwargs = {
            "sample_name": self.sample_name,
            "feature_name": self.feature_name,
        }

        # Initialize transformers
        # 1 | Center, scale and weigh the data
        scaler_kwargs = {
            "with_center": self.with_center,
            "with_std": self.with_std,
            "with_coslat": self.with_coslat,
        }
        self.scaler = GenericListTransformer(
            Scaler, compute=self.compute, **scaler_kwargs
        )
        # 2 | Rename dimensions
        self.renamer = GenericListTransformer(DimensionRenamer)
        # 3 | Convert MultiIndexes (before stacking)
        self.preconverter = GenericListTransformer(MultiIndexConverter)
        # 4 | Stack the data to 2D DataArray
        self.stacker = GenericListTransformer(Stacker, **dim_names_as_kwargs)
        # 5 | Convert MultiIndexes (after stacking)
        self.postconverter = GenericListTransformer(MultiIndexConverter)
        # 6 | Remove NaNs
        self.sanitizer = GenericListTransformer(
            Sanitizer, check_nans=self.check_nans, **dim_names_as_kwargs
        )
        # 7 | Concatenate into one 2D DataArray
        self.concatenator = Concatenator(**dim_names_as_kwargs)

    def get_serialization_attrs(self) -> dict:
        return dict(n_data=self.n_data)

    def transformer_types(self):
        """Ordered list of transformer operations."""
        return dict(
            scaler=Scaler,
            renamer=DimensionRenamer,
            preconverter=MultiIndexConverter,
            stacker=Stacker,
            postconverter=MultiIndexConverter,
            sanitizer=Sanitizer,
            concatenator=Concatenator,
        )

    def get_transformers(self, inverse: bool = False):
        transformers = [getattr(self, t) for t in self.transformer_types().keys()]
        if inverse:
            transformers = transformers[::-1]
        return transformers

    def fit(
        self,
        X: list[Data] | Data,
        sample_dims: Dims,
        weights: list[Data] | Data | None = None,
    ) -> Self:
        """Fit the preprocessor to the data.

        Parameters
        ----------
        X : xarray objects or list of xarray objects
            Input data.
        sample_dims : tuple of str
            Sample dimensions.
        weights : xr.DataArray or list of xr.DataArray, optional
            Weights to be applied to the data.

        Returns
        -------
        self : Preprocessor
            The fitted preprocessor.

        """
        self, X = self._fit_algorithm(X, sample_dims, weights)
        return self

    def _fit_algorithm(
        self,
        X: list[Data] | Data,
        sample_dims: Dims,
        weights: list[Data] | Data | None = None,
    ) -> tuple[Self, Data]:
        self._set_return_list(X)
        X = convert_to_list(X)
        self.n_data = len(X)
        sample_dims, feature_dims = get_dims(X, sample_dims)

        # For each DataArray a list of feature dimensions must be provided
        _check_parameter_number("feature_dims", feature_dims, self.n_data)

        # Ensure that weights are provided as a list
        weights = process_parameter("weights", weights, None, self.n_data)

        # 1 | Center, scale and weigh the data
        scaler_iterkwargs = {"weights": weights}
        X = self.scaler.fit_transform(
            X=X,
            sample_dims=sample_dims,
            feature_dims=feature_dims,
            iter_kwargs=scaler_iterkwargs,
        )
        # 2 | Rename dimensions
        X = self.renamer.fit_transform(X, sample_dims, feature_dims)
        sample_dims, feature_dims = extract_new_dim_names(self.renamer.transformers)
        # 3 | Convert MultiIndexes (before stacking)
        X = self.preconverter.fit_transform(X, sample_dims, feature_dims)
        # 4 | Stack the data to 2D DataArray
        X = self.stacker.fit_transform(X, sample_dims, feature_dims)
        # 5 | Convert MultiIndexes (after stacking)
        X = self.postconverter.fit_transform(X, sample_dims, feature_dims)
        # 6 | Remove NaNs
        X = self.sanitizer.fit_transform(X, sample_dims, feature_dims)
        # 7 | Concatenate into one 2D DataArray
        X = self.concatenator.fit_transform(X)  # type: ignore

        return self, X

    def transform(self, X: list[Data] | Data) -> DataArray:
        """Transform the data.

        Parameters
        ----------
        X : xarray objects or list of xarray objects
            Input data.

        Returns
        -------
        xr.DataArray
            The transformed data.

        """
        X = convert_to_list(X)

        if len(X) != self.n_data:
            raise ValueError(
                f"number of data objects passed should match number of data objects used for fitting"
                f"len(data objects)={len(X)} and "
                f"len(data objects used for fitting)={self.n_data}"
            )

        X_t = X.copy()
        for transformer in self.get_transformers():
            X_t = transformer.transform(X_t)  # type: ignore

        return X_t

    def fit_transform(
        self,
        X: list[Data] | Data,
        sample_dims: Dims,
        weights: list[Data] | Data | None = None,
    ) -> DataArray:
        # Take advantage of the fact that `.fit()` already transforms the data
        # to avoid duplicate computation
        self, X = self._fit_algorithm(X, sample_dims, weights)
        return X

    def inverse_transform_data(self, X: DataArray) -> list[Data] | Data:
        """Inverse transform the data.

        Parameters:
        -------------
        data: xr.DataArray
            Input data.

        Returns
        -------
        xr.DataArray or xr.Dataset or list of xr.DataArray
            The inverse transformed data.

        """
        X_it = X.copy()
        for transformer in self.get_transformers(inverse=True):
            X_it = transformer.inverse_transform_data(X_it)

        return self._process_output(X_it)

    def inverse_transform_components(self, X: DataArray) -> list[Data] | Data:
        """Inverse transform the components.

        Parameters:
        -------------
        data: xr.DataArray or list of xarray.DataArray
            Input data.

        Returns
        -------
        xr.DataArray
            The inverse transformed components.

        """
        X_it = X.copy()
        for transformer in self.get_transformers(inverse=True):
            X_it = transformer.inverse_transform_components(X_it)

        return self._process_output(X_it)

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        """Inverse transform the scores.

        This should be used for scores obtained from the fitted data.

        Parameters:
        -------------
        X: xr.DataArray
            Input data.

        Returns
        -------
        xr.DataArray
            The inverse transformed scores.

        """
        X_it = X.copy()
        for transformer in self.get_transformers(inverse=True):
            X_it = transformer.inverse_transform_scores(X_it)

        return X_it

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        """Inverse transform the scores.

        This should be used for scores obtained from new data.

        Parameters:
        -------------
        X: xr.DataArray
            Input data.

        Returns
        -------
        xr.DataArray
            The inverse transformed scores.

        """
        X_it = X.copy()
        for transformer in self.get_transformers(inverse=True):
            X_it = transformer.inverse_transform_scores_unseen(X_it)

        return X_it

    def _process_output(self, X: list[Data]) -> list[Data] | Data:
        if self.return_list:
            return X
        else:
            return unwrap_singleton_list(X)

    def _set_return_list(self, X):
        if isinstance(X, (list, tuple)):
            self.return_list = True
        else:
            self.return_list = False

    def serialize(self) -> xr.DataTree:
        """Serialize the necessary attributes of the fitted pre-processor
        and all transformers to a Dataset."""
        # Serialize the preprocessor as the root node
        dt = self._serialize()
        dt.name = "preprocessor"

        # Serialize all transformers
        names = list(self.transformer_types().keys())
        transformers = self.get_transformers()

        for name, transformer_obj in zip(names, transformers):
            dt_transformer = xr.DataTree()
            if isinstance(transformer_obj, GenericListTransformer):
                dt_transformer["transformers"] = xr.DataTree()
                # Loop through list transformer objects and assign a dummy key
                for i, transformer in enumerate(transformer_obj.transformers):
                    dt_transformer.transformers[str(i)] = transformer.serialize()
            else:
                dt_transformer = transformer_obj.serialize()
            # Place the serialized transformer in the tree
            dt[name] = dt_transformer

        return dt

    @classmethod
    def deserialize(cls, dt: xr.DataTree) -> Self:
        """Deserialize from a DataTree representation of the preprocessor
        and all attached Transformers."""
        # Create the parent preprocessor
        preprocessor = cls._deserialize(dt)

        # Loop through all transformers and deserialize
        names = list(preprocessor.transformer_types().keys())
        transformers = preprocessor.get_transformers()

        for name, transformer_obj in zip(names, transformers):
            if isinstance(transformer_obj, GenericListTransformer):
                # Recreate list transformers sequentially
                for transformer in dt[name].transformers.values():
                    deserialized = preprocessor.transformer_types()[name].deserialize(
                        transformer
                    )
                    transformer_obj.transformers.append(deserialized)
            else:
                # Recreate single transformer
                deserialized = preprocessor.transformer_types()[name].deserialize(
                    dt[name]
                )
                setattr(preprocessor, name, deserialized)

        return preprocessor



"""
Varimax-rotated Multivariate EOF analysis
============================================

Multivariate EOF analysis with additional Varimax rotation.
"""

# Load packages and data:
import matplotlib.pyplot as plt
import xarray as xr
from cartopy.crs import PlateCarree
from matplotlib.gridspec import GridSpec

import xeofs as xe

# %%
# Create four different dataarrayss
sst = xr.tutorial.open_dataset("ersstv5")["sst"]
subset1 = sst.isel(lon=slice(0, 45))
subset2 = sst.isel(lon=slice(46, 90))
subset3 = sst.isel(lon=slice(91, 135))
subset4 = sst.isel(lon=slice(136, None))

# %%
# Perform the actual analysis

multivariate_data = [subset1, subset2, subset3, subset4]
mpca = xe.single.EOF(n_modes=100, standardize=False, use_coslat=True)
mpca.fit(multivariate_data, dim="time")
rotator = xe.single.EOFRotator(n_modes=20)
rotator.fit(mpca)
rcomponents = rotator.components()
rscores = rotator.scores()

# %%
# Plot mode 1

mode = 5
proj = PlateCarree()
kwargs = {
    "cmap": "RdBu",
    "vmin": -0.05,
    "vmax": 0.05,
    "transform": proj,
    "add_colorbar": False,
}

fig = plt.figure(figsize=(7.3, 6))
fig.subplots_adjust(wspace=0)
gs = GridSpec(2, 4, figure=fig, width_ratios=[1, 1, 1, 1])
ax = [fig.add_subplot(gs[0, i], projection=proj) for i in range(4)]
ax_pc = fig.add_subplot(gs[1, :])

# PC
rscores.sel(mode=mode).plot(ax=ax_pc)
ax_pc.set_xlabel("")
ax_pc.set_title("")

# EOFs
for i, (a, comps) in enumerate(zip(ax, rcomponents)):
    a.coastlines(color=".5")
    comps.sel(mode=mode).plot(ax=a, **kwargs)
    a.set_xticks([], [])
    a.set_yticks([], [])
    a.set_xlabel("")
    a.set_ylabel("")
    a.set_title("Subset {:}".format(i + 1))
ax[0].set_ylabel("EOFs")
fig.suptitle("Mode {:}".format(mode))
plt.savefig("mreof-analysis.jpg")



import warnings

import numpy as np
import xarray as xr
from typing_extensions import Self

from ..linalg import total_variance
from ..preprocessing import PCA
from ..utils.data_types import DataArray, DataObject
from ..utils.xarray_utils import argsort_dask
from .base_model_single_set import BaseModelSingleSet


class POP(BaseModelSingleSet):
    """Principal Oscillation Pattern (POP) analysis.

    POP analysis [1]_ [2]_ is a linear multivariate technique used to identify
    and describe dominant oscillatory modes in a dynamical system. POP analysis
    involves computing the eigenvalues and eigenvectors of the `feedback matrix`
    defined as

    .. math::
        A = C_1 C_0^{-1}

    where :math:`C_0` is the covariance matrix and :math:`C_1` is the lag-1
    covariance matrix of the input data. The eigenvectors of the feedback matrix
    are the POPs and the eigenvalues are related to the damping times and
    periods of the oscillatory modes.

    Parameters
    ----------
    n_modes: int, default=10
        Number of modes to calculate.
    center: bool, default=True
        Whether to center the input data.
    standardize: bool, default=False
        Whether to standardize the input data.
    use_coslat: bool, default=False
        Whether to use cosine of latitude for scaling.
    use_pca : bool, default=False
        If True, perform PCA to reduce the dimensionality of the data.
    n_pca_modes : int | float | str, default=0.999
        If int, specifies the number of modes to retain. If float, specifies the
        fraction of variance in the (whitened) data that should be explained by
        the retained modes. If "all", all modes are retained.
    init_rank_reduction : float, default=0.3
        Only relevant when `use_pca=True` and `n_modes` is a float, in which
        case it denotes the fraction of the initial rank to reduce the data to
        via PCA as a first guess before truncating the solution to the desired
        fraction of explained variance. This allows for faster computation of
        PCA via randomized SVD and avoids the need to compute the full SVD.
    sample_name: str, default="sample"
        Name of the sample dimension.
    feature_name: str, default="feature"
        Name of the feature dimension.
    check_nans : bool, default=True
        If True, remove full-dimensional NaN features from the data, check to
        ensure that NaN features match the original fit data during transform,
        and check for isolated NaNs. Note: this forces eager computation of dask
        arrays. If False, skip all NaN checks. In this case, NaNs should be
        explicitly removed or filled prior to fitting, or SVD will fail.
    compute : bool, default=True
        Whether to compute elements of the model eagerly, or to defer computation.
        If True, four pieces of the fit will be computed sequentially: 1) the
        preprocessor scaler, 2) optional NaN checks, 3) SVD decomposition, 4) scores
        and components.
    random_state : int, optional
        Seed for the random number generator.
    solver: {"auto", "full", "randomized"}, default="auto"
        Solver to use for the SVD computation.
    solver_kwargs: dict, default={}
        Additional keyword arguments to be passed to the SVD solver.

    References
    ----------
    .. [1] Hasselmann, K. PIPs and POPs: The reduction of complex dynamical systems using principal interaction and oscillation patterns. J. Geophys. Res. 93, 11015–11021 (1988).
    .. [2] von Storch, H., G. Bürger, R. Schnur, and J. von Storch, 1995:
    Principal Oscillation Patterns: A Review. J. Climate, 8, 377–400,
    https://doi.org/10.1175/1520-0442(1995)008<0377:POPAR>2.0.CO;2.


    Examples
    --------

    Perform POP analysis in PC space spanned by the first 10 modes:

    >>> pop = xe.single.POP(n_modes="all", use_pca=True, n_pca_modes=10)
    >>> pop.fit(X, "time)

    Get the POPs and associated time coefficients:

    >>> patterns = pop.components()
    >>> scores = pop.scores()

    Reconstruct the original data using a conjugate pair of POPs:

    >>> pop_pairs = scores.sel(mode=[1, 2])
    >>> X_rec = pop.inverse_transform(pop_pairs)

    """

    uses_complex = True

    def __init__(
        self,
        n_modes: int = 2,
        center: bool = True,
        standardize: bool = False,
        use_coslat: bool = False,
        use_pca: bool = True,
        n_pca_modes: float | int = 0.999,
        pca_init_rank_reduction: float = 0.3,
        check_nans=True,
        sample_name: str = "sample",
        feature_name: str = "feature",
        compute: bool = True,
        random_state: int | None = None,
        solver: str = "auto",
        solver_kwargs: dict = {},
        **kwargs,
    ):
        super().__init__(
            n_modes=n_modes,
            center=center,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            sample_name=sample_name,
            feature_name=feature_name,
            compute=compute,
            random_state=random_state,
            solver=solver,
            solver_kwargs=solver_kwargs,
            **kwargs,
        )
        self.attrs.update({"model": "Principal Oscillation Pattern analysis"})

        self.pca = PCA(
            use_pca=use_pca,
            n_modes=n_pca_modes,
            init_rank_reduction=pca_init_rank_reduction,
            sample_name=sample_name,
            feature_name=feature_name,
            compute_eagerly=compute,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
        )

        self.sorted = False

    def get_serialization_attrs(self) -> dict:
        return dict(
            data=self.data,
            preprocessor=self.preprocessor,
            pca=self.pca,
            sorted=self.sorted,
        )

    def _np_solve_pop_system(self, X):
        # Feedack matrix
        A = X[1:].conj().T @ X[:-1] @ np.linalg.inv(X[:-1].conj().T @ X[:-1])

        # Compute POPs
        lbda, P = np.linalg.eig(A)

        # e-folding times /damping times
        tau = -1 / np.log(abs(lbda))

        # POP periods
        with warnings.catch_warnings(record=True):
            warnings.filterwarnings(
                "ignore", "divide by zero encountered", RuntimeWarning
            )

            T = 2 * np.pi / np.angle(lbda)

        # POP (time) coefficients (Storch et al. 1995, equation 19)
        Z = self._np_compute_pop_coefficients(X, P)
        # Reconstruction of original data
        # Xrec = Z @ P.T
        # It seems that the signs of some columns of Xrec are flipped, probably

        return P, Z, lbda, T, tau

    def _np_compute_pop_coefficients(self, X, P):
        # POP (time) coefficients (Storch et al. 1995, equation 19)
        Z = np.empty((X.shape[0], P.shape[1]), dtype=complex)
        for i in range(P.shape[1]):
            p = P[:, i : i + 1]
            pr = p.real
            pi = p.imag

            M = np.array([[pr.T @ pr, pr.T @ pi], [pr.T @ pi, pi.T @ pi]]).squeeze()
            Minv = np.linalg.pinv(M)
            zri = Minv @ np.hstack([X @ pr, X @ pi]).T
            z = zri[0] + 1j * zri[1]
            Z[:, i] = z
        return Z

    def _fit_algorithm(self, X: DataArray) -> Self:
        sample_name = self.sample_name
        feature_name = self.feature_name

        # Transform in PC space
        X = self.pca.fit_transform(X)

        P, Z, lbda, T, tau = xr.apply_ufunc(
            self._np_solve_pop_system,
            X,
            input_core_dims=[[sample_name, feature_name]],
            output_core_dims=[
                [feature_name, "mode"],
                [sample_name, "mode"],
                ["mode"],
                ["mode"],
                ["mode"],
            ],
            dask="allowed",
        )

        mode_coords = np.arange(1, P.mode.size + 1)
        P = P.assign_coords(mode=mode_coords)
        Z = Z.assign_coords(mode=mode_coords)
        lbda = lbda.assign_coords(mode=mode_coords)
        T = T.assign_coords(mode=mode_coords)
        tau = tau.assign_coords(mode=mode_coords)

        # Compute dynamical importance of each mode
        var_Z = Z.var(sample_name)
        norms = (var_Z) ** (0.5)

        # Compute total variance
        var_tot = total_variance(X, sample_name)

        # Reorder according to variance
        idx_modes_sorted = argsort_dask(norms, "mode")[::-1]  # type: ignore
        idx_modes_sorted.coords.update(norms.coords)

        P = self.pca.inverse_transform_components(P)

        # Store the results
        self.data.add(X, "input_data", allow_compute=False)
        self.data.add(P, "components")
        self.data.add(Z, "scores")
        self.data.add(norms, "norms")
        self.data.add(lbda, "eigenvalues")
        self.data.add(tau, "damping_times")
        self.data.add(T, "periods")
        self.data.add(idx_modes_sorted, "idx_modes_sorted")
        self.data.add(var_tot, "total_variance")

        self.data.set_attrs(self.attrs)
        return self

    def _post_compute(self):
        """Leave sorting until after compute because it can't be done lazily."""
        self._sort_by_variance()

    def _sort_by_variance(self):
        """Re-sort the mode dimension of all data variables by variance explained."""
        if not self.sorted:
            for key in self.data.keys():
                if "mode" in self.data[key].dims and key != "idx_modes_sorted":
                    self.data[key] = (
                        self.data[key]
                        .isel(mode=self.data["idx_modes_sorted"].values)
                        .assign_coords(mode=self.data[key].mode)
                    )
        self.sorted = True

    def _transform_algorithm(self, X: DataArray) -> DataArray:
        sample_name = self.sample_name
        feature_name = self.feature_name

        P = self.data["components"]

        # Transform into PC spcae
        P = self.pca.transform_components(P)
        X = self.pca.transform(X)

        # Project the data
        Z = xr.apply_ufunc(
            self._np_compute_pop_coefficients,
            X,
            P,
            input_core_dims=[[sample_name, feature_name], [feature_name, "mode"]],
            output_core_dims=[[sample_name, "mode"]],
            dask="allowed",
        )
        Z.name = "scores"

        Z = self.pca.inverse_transform_scores(Z)

        return Z

    def _inverse_transform_algorithm(self, scores: DataArray) -> DataArray:
        """Reconstruct the original data from transformed data.

        Parameters
        ----------
        scores: DataArray
            Transformed data to be reconstructed. This could be a subset
            of the `scores` data of a fitted model, or unseen data. Must
            have a 'mode' dimension.

        Returns
        -------
        data: DataObject
            Reconstructed data.

        """
        # Reconstruct the data
        P = self.data["components"].sel(mode=scores.mode)

        # Transform in PC space
        P = self.pca.transform_components(P)

        reconstructed_data = xr.dot(scores, P, dims="mode")
        reconstructed_data.name = "reconstructed_data"

        # Inverse transform the data into physical space
        reconstructed_data = self.pca.inverse_transform_data(reconstructed_data)

        return reconstructed_data

    def components(self) -> DataObject:
        """Return the POPs.

        The POPs are the eigenvectors of the feedback matrix.

        Returns
        -------
        components: DataObject
            Principal Oscillation Patterns (POPs).

        """
        return super().components(normalized=False)

    def scores(self, normalized: bool = False) -> DataArray:
        """Return the POP coefficients/scores.

        Parameters
        ----------
        normalized : bool, default=True
            Whether to normalize the scores by the L2 norm.

        Returns
        -------
        components: DataObject
            POP coefficients.

        """
        return super().scores(normalized=normalized)

    def eigenvalues(self) -> DataArray:
        """Return the eigenvalues of the feedback matrix.

        Returns
        -------
        DataArray
            Real or complex eigenvalues.

        """
        return self.data["eigenvalues"]

    def damping_times(self) -> DataArray:
        """Return the damping times of the feedback matrix.

        The damping times are defined as

        .. math::
            \\tau = -\\frac{1}{\\log(|\\lambda|)}

        where :math:`\\lambda` is the eigenvalue.

        Returns
        -------
        DataArray
            Damping times.

        """
        return self.data["damping_times"]

    def periods(self) -> DataArray:
        """Return the periods of the feedback matrix.

        For complex eigenvalues, the periods are defined as

        .. math::
            T = \\frac{2\\pi}{\\arg(\\lambda)}

        where :math:`\\lambda` is the eigenvalue. For real eigenvalues ``inf``
        is returned.

        Returns
        -------
        DataArray
            Periods.

        """
        return self.data["periods"]

    def components_amplitude(self) -> DataObject:
        """Return the amplitude of the POP components.

        The amplitude of the components are defined as

        .. math::
            A_{ij} = |C_{ij}|

        where :math:`C_{ij}` is the :math:`i`-th entry of the :math:`j`-th component and
        :math:`|\\cdot|` denotes the absolute value.


        Returns
        -------
        components_amplitude: DataObject
            Amplitude of the components of the fitted model.

        """
        amplitudes = abs(self.data["components"])

        amplitudes.name = "components_amplitude"
        return self.preprocessor.inverse_transform_components(amplitudes)

    def components_phase(self) -> DataObject:
        """Return the phase of the POP components.

        The phase of the components are defined as

        .. math::
            \\phi_{ij} = \\arg(C_{ij})

        where :math:`C_{ij}` is the :math:`i`-th entry of the :math:`j`-th component and
        :math:`\\arg(\\cdot)` denotes the argument of a complex number.

        Returns
        -------
        components_phase: DataObject
            Phase of the components of the fitted model.

        """
        comps = self.data["components"]
        comp_phase = xr.apply_ufunc(np.angle, comps, dask="allowed", keep_attrs=True)
        comp_phase.name = "components_phase"
        return self.preprocessor.inverse_transform_components(comp_phase)

    def scores_amplitude(self, normalized=True) -> DataArray:
        """Return the amplitude of the POP coefficients/scores.

        The amplitude of the scores are defined as

        .. math::
            A_{ij} = |S_{ij}|

        where :math:`S_{ij}` is the :math:`i`-th entry of the :math:`j`-th score and
        :math:`|\\cdot|` denotes the absolute value.

        Parameters
        ----------
        normalized : bool, default=True
            Whether to normalize the scores by the singular values.

        Returns
        -------
        scores_amplitude: DataObject
            Amplitude of the scores of the fitted model.

        """
        scores = self.data["scores"].copy()
        if normalized:
            scores = scores / self.data["norms"]

        amplitudes = abs(scores)
        amplitudes.name = "scores_amplitude"
        return self.preprocessor.inverse_transform_scores(amplitudes)

    def scores_phase(self) -> DataArray:
        """Return the phase of the POP coefficients/scores.

        The phase of the scores are defined as

        .. math::
            \\phi_{ij} = \\arg(S_{ij})

        where :math:`S_{ij}` is the :math:`i`-th entry of the :math:`j`-th score and
        :math:`\\arg(\\cdot)` denotes the argument of a complex number.

        Returns
        -------
        scores_phase: DataObject
            Phase of the scores of the fitted model.

        """
        scores = self.data["scores"]
        phases = xr.apply_ufunc(np.angle, scores, dask="allowed", keep_attrs=True)
        phases.name = "scores_phase"
        return self.preprocessor.inverse_transform_scores(phases)



import xarray as xr
import pytest

from xeofs.utils.dimension_renamer import DimensionRenamer

import pandas as pd


@pytest.fixture
def da_simple():
    # Basic DataArray with single dimension
    return xr.DataArray([1, 2, 3], coords={"dim0": ["a", "b", "c"]}, dims=["dim0"])


@pytest.fixture
def da_multi():
    # DataArray with MultiIndex
    arrays = [
        pd.Index(["a", "b", "c"], name="dim1a"),
        pd.Index([1, 2, 3], name="dim1b"),
    ]
    multi_index = pd.MultiIndex.from_arrays(arrays)
    return xr.DataArray([1, 2, 3], coords={"dim1": multi_index}, dims=["dim1"])


def test_simple_dim_rename(da_simple):
    renamer = DimensionRenamer("dim0", "_suffix")
    da_new = renamer.fit_transform(da_simple)
    assert "dim0_suffix" in da_new.dims

    # Inverse transform
    da_orig = renamer.inverse_transform(da_new)
    assert "dim0" in da_orig.dims


def test_multiindex_dim_rename(da_multi):
    renamer = DimensionRenamer("dim1", "_suffix")
    da_new = renamer.fit_transform(da_multi)
    assert "dim1_suffix" in da_new.dims
    assert "dim1a_suffix" in da_new.coords["dim1_suffix"].coords.keys()
    assert "dim1b_suffix" in da_new.coords["dim1_suffix"].coords.keys()

    # Inverse transform
    da_orig = renamer.inverse_transform(da_new)
    assert "dim1" in da_orig.dims
    assert "dim1a" in da_orig.coords["dim1"].coords.keys()
    assert "dim1b" in da_orig.coords["dim1"].coords.keys()


def test_fit_without_transform(da_simple):
    renamer = DimensionRenamer("dim0", "_suffix")
    renamer.fit(da_simple)
    assert hasattr(renamer, "dims_mapping")
    assert renamer.dims_mapping == {"dim0": "dim0_suffix"}


def test_incorrect_dim_name(da_simple):
    with pytest.raises(KeyError):
        renamer = DimensionRenamer("nonexistent_dim", "_suffix")
        renamer.fit_transform(da_simple)


def test_empty_suffix(da_simple):
    renamer = DimensionRenamer("dim0", "")
    da_new = renamer.fit_transform(da_simple)
    assert "dim0" in da_new.dims



import warnings
from typing import Sequence

import numpy as np

from ..utils.data_types import DataArray
from .cpcca import CPCCA, ComplexCPCCA, HilbertCPCCA


class MCA(CPCCA):
    """Maximum Covariance Analysis (MCA).

    MCA seeks to find paris of coupled patterns that maximize the squared
    covariance [1]_ [2]_ .

    This method solves the following optimization problem:

        :math:`\\max_{q_x, q_y} \\left( q_x^T X^T Y q_y \\right)`

    subject to the constraints:

        :math:`q_x^T q_x = 1, \\quad q_y^T q_y = 1`

    where :math:`X` and :math:`Y` are the input data matrices and :math:`q_x`
    and :math:`q_y` are the corresponding pattern vectors.

    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    standardize : Squence[bool] | bool, default=False
        Whether to standardize the input data. Generally not recommended as
        standardization can be managed by the degree of whitening.
    use_coslat : Sequence[bool] | bool, default=False
        For data on a longitude-latitude grid, whether to correct for varying
        grid cell areas towards the poles by scaling each grid point with the
        square root of the cosine of its latitude.
    use_pca : Sequence[bool] | bool, default=False
        Whether to preprocess each field individually by reducing dimensionality
        through PCA. The cross-covariance matrix is then computed in the reduced
        principal component space.
    n_pca_modes : Sequence[int | float | str] | int | float | str, default=0.999
        Number of modes to retain during PCA preprocessing step. If int,
        specifies the exact number of modes; if float, specifies the fraction of
        variance to retain; if "all", all modes are retained.
    pca_init_rank_reduction : Sequence[float] | float, default=0.3
        Relevant when `use_pca=True` and `n_pca_modes` is a float. Specifies the
        initial fraction of rank reduction for faster PCA computation via
        randomized SVD.
    check_nans : Sequence[bool] | bool, default=True
        Whether to check for NaNs in the input data. Set to False for lazy model
        evaluation.
    compute : bool, default=True
        Whether to compute the model elements eagerly. If True, the following
        are computed sequentially: preprocessor scaler, optional NaN checks, SVD
        decomposition, scores, and components.
    random_state : numpy.random.Generator | int | None, default=None
        Seed for the random number generator.
    sample_name : str, default="sample"
        Name for the new sample dimension.
    feature_name : Sequence[str] | str, default="feature"
        Name for the new feature dimension.
    solver : {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver function.


    References
    ----------
    .. [1] Bretherton, C., Smith, C., Wallace, J., 1992. An intercomparison of
        methods for finding coupled patterns in climate data. Journal of climate
        5, 541–560.
    .. [2] Wilks, D. S. Statistical Methods in the Atmospheric Sciences.
        (Academic Press, 2019).
        doi:https://doi.org/10.1016/B978-0-12-815823-4.00011-0.

    Examples
    --------

    Perform MCA on two datasets on a regular longitude-latitude grid:

    >>> model = MCA(n_modes=5, use_coslat=True)
    >>> model.fit(X, Y, dim="time")

    """

    def __init__(
        self,
        n_modes: int = 2,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        check_nans: Sequence[bool] | bool = True,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
    ):
        CPCCA.__init__(
            self,
            n_modes=n_modes,
            alpha=[1.0, 1.0],
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            use_pca=use_pca,
            n_pca_modes=n_pca_modes,
            pca_init_rank_reduction=pca_init_rank_reduction,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
        )
        self.attrs.update({"model": "Maximum Covariance Analysis"})
        # Renove alpha from the inherited CPCCA serialization params because it is hard-coded for MCA
        self._params.pop("alpha")

    def covariance_fraction_CD95(self):
        """Get the covariance fraction (CF).

        Cheng and Dunkerton (1995) [3]_ define the CF as follows:

        .. math::
            CF_i = \\frac{\\sigma_i}{\\sum_{i=1}^{m} \\sigma_i}

        where `m` is the total number of modes and :math:`\\sigma_i` is the
        `ith` singular value of the covariance matrix.

        This implementation estimates the sum of singular values from the first
        `n` modes, therefore one should aim to retain as many modes as possible
        to get a good estimate of the covariance fraction.

        Note
        ----
        In MCA, the focus is on maximizing the *squared* covariance (SC). As a
        result, this quantity is preserved during decomposition - meaning the SC
        of both datasets remains unchanged before and after decomposition. Each
        mode explains a fraction of the total SC, and together, all modes can
        reconstruct the total SC of the cross-covariance matrix. However, the
        (non-squared) covariance is not invariant in MCA; it is not preserved by
        the individual modes and cannot be reconstructed from them.
        Consequently, the squared covariance fraction (SCF) is invariant in MCA
        and is typically used to assess the relative importance of each mode. In
        contrast, the convariance fraction (CF) is not invariant. Cheng and
        Dunkerton [3]_ introduced the CF to compare the relative importance of
        modes before and after Varimax rotation in MCA. Notably, when the data
        fields in MCA are identical, the CF corresponds to the explained
        variance ratio in Principal Component Analysis (PCA).

        References
        ----------
        .. [3] Cheng, X. & Dunkerton, T. J. Orthogonal Rotation of Spatial
            Patterns Derived from Singular Value Decomposition Analysis. J.
            Climate 8, 2631–2643 (1995).

        """
        # Check how sensitive the CF is to the number of modes
        cov_exp = self._covariance_explained_DC95()
        tot_var = self._total_covariance()
        cf = cov_exp[0] / cov_exp.cumsum()
        change_per_mode = cf.shift({"mode": 1}) - cf
        change_in_cf_in_last_mode = change_per_mode.isel(mode=-1)
        if change_in_cf_in_last_mode > 0.001:
            warnings.warn(
                "The curent estimate of CF is sensitive to the number of modes retained. Please increase `n_modes` for a better estimate."
            )
        cov_frac = cov_exp / tot_var
        cov_frac.name = "covariance_fraction"
        cov_frac.attrs.update(cov_exp.attrs)
        return cov_frac

    def _squared_covariance(self) -> DataArray:
        """Get the squared covariance.

        The squared covariance is given by the squared singular values of the
        covariance matrix:

        .. math::
            SC_i = \\sigma_i^2

        where :math:`\\sigma_i` is the `ith` singular value of the covariance
        matrix.

        """
        # only true for MCA, for alpha < 1 the sigmas become more and more correlation coefficients
        # either remove this one and provide it only for MCA child class, or use error formulation
        sc = self.data["squared_covariance"]
        sc.name = "squared_covariance"
        return sc

    def _covariance_explained_DC95(self) -> DataArray:
        """Get the covariance explained (CE) per mode according to CD95.

        References
        ----------
        Cheng, X. & Dunkerton, T. J. Orthogonal Rotation of Spatial Patterns
        Derived from Singular Value Decomposition Analysis. J. Climate 8,
        2631–2643 (1995).

        """
        cov_exp = self._squared_covariance() ** (0.5)
        cov_exp.name = "pseudo_explained_covariance"
        return cov_exp

    def _total_covariance(self) -> DataArray:
        """Get the total covariance.

        This measure follows the defintion of Cheng and Dunkerton (1995).
        Note that this measure is not an invariant in MCA.

        """
        pseudo_tot_cov = self._covariance_explained_DC95().sum()
        pseudo_tot_cov.name = "pseudo_total_covariance"
        return pseudo_tot_cov


class ComplexMCA(ComplexCPCCA, MCA):
    """Complex MCA.

    MCA applied to a complex-valued field obtained from a pair of variables such
    as the zonal and meridional components, :math:`U` and :math:`V`, of the wind
    field. Complex EOF analysis then maximizes the squared covariance between
    two datasets of the form

    .. math::
        Z_x = U_x + iV_x

    and

    .. math::
        Z_y = U_y + iV_y

    into a set of complex-valued components and PC scores.


    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    standardize : Squence[bool] | bool, default=False
        Whether to standardize the input data. Generally not recommended as
        standardization can be managed by the degree of whitening.
    use_coslat : Sequence[bool] | bool, default=False
        For data on a longitude-latitude grid, whether to correct for varying
        grid cell areas towards the poles by scaling each grid point with the
        square root of the cosine of its latitude.
    use_pca : Sequence[bool] | bool, default=False
        Whether to preprocess each field individually by reducing dimensionality
        through PCA. The cross-covariance matrix is then computed in the reduced
        principal component space.
    n_pca_modes : Sequence[int | float | str] | int | float | str, default=0.999
        Number of modes to retain during PCA preprocessing step. If int,
        specifies the exact number of modes; if float, specifies the fraction of
        variance to retain; if "all", all modes are retained.
    pca_init_rank_reduction : Sequence[float] | float, default=0.3
        Relevant when `use_pca=True` and `n_pca_modes` is a float. Specifies the
        initial fraction of rank reduction for faster PCA computation via
        randomized SVD.
    check_nans : Sequence[bool] | bool, default=True
        Whether to check for NaNs in the input data. Set to False for lazy model
        evaluation.
    compute : bool, default=True
        Whether to compute the model elements eagerly. If True, the following
        are computed sequentially: preprocessor scaler, optional NaN checks, SVD
        decomposition, scores, and components.
    random_state : numpy.random.Generator | int | None, default=None
        Seed for the random number generator.
    sample_name : str, default="sample"
        Name for the new sample dimension.
    feature_name : Sequence[str] | str, default="feature"
        Name for the new feature dimension.
    solver : {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver function.

    Examples
    --------

    With two DataArrays `u_i` and `v_i` representing the zonal and meridional
    components of the wind field for two different regions :math:`x` and
    :math:`y`, construct

    >>> X = u_x + 1j * v_x
    >>> Y = u_y + 1j * v_y

    and fit the Complex MCA model:

    >>> model = ComplexMCA(n_modes=5)
    >>> model.fit(X, Y, "time")


    """

    def __init__(
        self,
        n_modes: int = 2,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        check_nans: Sequence[bool] | bool = True,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
    ):
        ComplexCPCCA.__init__(
            self,
            n_modes=n_modes,
            alpha=[1.0, 1.0],
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            use_pca=use_pca,
            n_pca_modes=n_pca_modes,
            pca_init_rank_reduction=pca_init_rank_reduction,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
        )
        self.attrs.update({"model": "Complex MCA"})
        # Renove alpha from the inherited CPCCA serialization params because it is hard-coded for MCA
        self._params.pop("alpha")


class HilbertMCA(HilbertCPCCA, ComplexMCA):
    """Hilbert MCA.

    Hilbert MCA [1]_ (aka Analytical SVD),  extends MCA by
    examining amplitude-phase relationships. It augments the input data with its
    Hilbert transform, creating a complex-valued field.

    This method solves the following optimization problem:

        :math:`\\max_{q_x, q_y} \\left( q_x^H X^H Y q_y \\right)`

    subject to the constraints:

        :math:`q_x^H q_x = 1, \\quad q_y^H q_y = 1`

    where :math:`H` denotes the conjugate transpose and :math:`X` and :math:`Y`
    are the augmented data matrices.

    An optional padding with exponentially decaying values can be applied prior
    to the Hilbert transform in order to mitigate the impact of spectral
    leakage.


    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    padding : Sequence[str] | str | None, default="exp"
        Padding method for the Hilbert transform. Available options are: - None:
        no padding - "exp": exponential decay
    decay_factor : Sequence[float] | float, default=0.2
        Decay factor for the exponential padding.
    standardize : Squence[bool] | bool, default=False
        Whether to standardize the input data. Generally not recommended as
        standardization can be managed by the degree of whitening.
    use_coslat : Sequence[bool] | bool, default=False
        For data on a longitude-latitude grid, whether to correct for varying
        grid cell areas towards the poles by scaling each grid point with the
        square root of the cosine of its latitude.
    use_pca : Sequence[bool] | bool, default=False
        Whether to preprocess each field individually by reducing dimensionality
        through PCA. The cross-covariance matrix is computed in the reduced
        principal component space.
    n_pca_modes : Sequence[int | float | str] | int | float | str, default=0.999
        Number of modes to retain during PCA preprocessing step. If int,
        specifies the exact number of modes; if float, specifies the fraction of
        variance to retain; if "all", all modes are retained.
    pca_init_rank_reduction : Sequence[float] | float, default=0.3
        Relevant when `use_pca=True` and `n_pca_modes` is a float. Specifies the
        initial fraction of rank reduction for faster PCA computation via
        randomized SVD.
    check_nans : Sequence[bool] | bool, default=True
        Whether to check for NaNs in the input data. Set to False for lazy model
        evaluation.
    compute : bool, default=True
        Whether to compute the model elements eagerly. If True, the following
        are computed sequentially: preprocessor scaler, optional NaN checks, SVD
        decomposition, scores, and components.
    random_state : numpy.random.Generator | int | None, default=None
        Seed for the random number generator.
    sample_name : str, default="sample"
        Name for the new sample dimension.
    feature_name : Sequence[str] | str, default="feature"
        Name for the new feature dimension.
    solver : {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver function.

    References
    ----------
    .. [1] Elipot, S., Frajka-Williams, E., Hughes, C. W., Olhede, S. &
        Lankhorst, M. Observed Basin-Scale Response of the North Atlantic
        Meridional Overturning Circulation to Wind Stress Forcing. Journal of
        Climate 30, 2029–2054 (2017).



    Examples
    --------
    >>> model = HilbertMCA(n_modes=5)
    >>> model.fit(X, Y, "time")

    """

    def __init__(
        self,
        n_modes: int = 2,
        padding: Sequence[str] | str | None = "exp",
        decay_factor: Sequence[float] | float = 0.2,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        check_nans: Sequence[bool] | bool = True,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
    ):
        HilbertCPCCA.__init__(
            self,
            n_modes=n_modes,
            alpha=[1.0, 1.0],
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            use_pca=use_pca,
            n_pca_modes=n_pca_modes,
            pca_init_rank_reduction=pca_init_rank_reduction,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
            padding=padding,
            decay_factor=decay_factor,
        )
        self.attrs.update({"model": "Hilbert MCA"})
        # Renove alpha from the inherited CPCCA serialization params because it is hard-coded for MCA
        self._params.pop("alpha")



from ..utils.data_types import DataArray


def total_variance(X: DataArray, dim: str) -> DataArray:
    """Compute the total variance of the centered data."""
    return (X * X.conj()).sum() / (X[dim].size - 1)



"""
Removing nonlinear trends with EEOF analysis
============================================

This tutorial illustrates the application of Extended EOF (EEOF) analysis
to isolate and remove nonlinear trends within a dataset.

Let's begin by setting up the required packages and fetching the data.
"""

import matplotlib.pyplot as plt
import xarray as xr

import xeofs as xe

xr.set_options(display_expand_data=False)

# %%
# We load the sea surface temperature (SST) data from the xarray tutorial.
# The dataset consists of monthly averages from 1970 to 2021. To ensure the seasonal
# cycle doesn't overshadow the analysis, we remove the monthly climatologies.

sst = xr.tutorial.open_dataset("ersstv5").sst
sst = sst.groupby("time.month") - sst.groupby("time.month").mean("time")


# %%
# We start by performing a standard EOF analysis on the dataset.

eof = xe.single.EOF(n_modes=10)
eof.fit(sst, dim="time")
scores = eof.scores()
components = eof.components()

# %%
# We immediately see that the first mode represents the global warming trend.
# Yet, the signal is somewhat muddled by short-term and year-to-year variations.
# Note the pronounced spikes around 1998 and 2016, hinting at the leakage of
# ENSO signatures into this mode.

fig, ax = plt.subplots(1, 2, figsize=(10, 5))
scores.sel(mode=1).plot(ax=ax[0])
components.sel(mode=1).plot(ax=ax[1])


# %%
# Now, let's try to identify this trend more cleanly. To this end, we perform an
# EEOF analysis on the same data with a suitably large embedding dimension.
# We choose an embedding dimensioncorresponding to 120 months which is large enough
# to capture long-term trends. To speed up computation, we apply the EEOF analysis
# to the extended (lag) covariance matrix derived from the first 50 PCs.

eeof = xe.single.ExtendedEOF(n_modes=5, tau=1, embedding=120, n_pca_modes=50)
eeof.fit(sst, dim="time")
components_ext = eeof.components()
scores_ext = eeof.scores()

# %%
# The first mode now represents the global warming trend much more clearly.

fig, ax = plt.subplots(1, 2, figsize=(10, 5))
scores_ext.sel(mode=1).plot(ax=ax[0])
components_ext.sel(mode=1, embedding=0).plot(ax=ax[1])

# %%
# We can use this to the first mode to remove this nonlinear trend from our original dataset.

sst_trends = eeof.inverse_transform(scores_ext.sel(mode=1))
sst_detrended = sst - sst_trends


# %%
# Reapplying the standard EOF analysis on our now detrended dataset:

eof_model_detrended = xe.single.EOF(n_modes=5)
eof_model_detrended.fit(sst_detrended, dim="time")
scores_detrended = eof_model_detrended.scores()
components_detrended = eof_model_detrended.components()


# %%
# The first mode now represents ENSO without any trend component.

fig, ax = plt.subplots(1, 2, figsize=(10, 5))
scores_detrended.sel(mode=1).plot(ax=ax[0])
components_detrended.sel(mode=1).plot(ax=ax[1])


# %%



import pytest
import xarray as xr
import numpy as np

from xeofs.preprocessing.scaler import Scaler


@pytest.mark.parametrize(
    "with_std, with_coslat",
    [
        (True, True),
        (True, True),
        (True, False),
        (True, False),
        (False, True),
        (False, True),
        (False, False),
        (False, False),
        (True, True),
        (True, True),
        (True, False),
        (True, False),
        (False, True),
        (False, True),
        (False, False),
        (False, False),
    ],
)
def test_init_params(with_std, with_coslat):
    s = Scaler(with_std=with_std, with_coslat=with_coslat)
    assert s.get_params()["with_std"] == with_std
    assert s.get_params()["with_coslat"] == with_coslat


@pytest.mark.parametrize(
    "with_std, with_coslat",
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_fit_params(with_std, with_coslat, mock_data_array):
    s = Scaler(with_std=with_std, with_coslat=with_coslat)
    sample_dims = ["time"]
    feature_dims = ["lat", "lon"]
    size_lats = mock_data_array.lat.size
    weights = xr.DataArray(np.random.rand(size_lats), dims=["lat"])
    s.fit(mock_data_array, sample_dims, feature_dims, weights)
    assert hasattr(s, "mean_"), "Scaler has no mean attribute."
    if with_std:
        assert hasattr(s, "std_"), "Scaler has no std attribute."
    if with_coslat:
        assert hasattr(s, "coslat_weights_"), "Scaler has no coslat_weights attribute."
    assert s.mean_ is not None, "Scaler mean is None."
    if with_std:
        assert s.std_ is not None, "Scaler std is None."
    if with_coslat:
        assert s.coslat_weights_ is not None, "Scaler coslat_weights is None."


@pytest.mark.parametrize(
    "with_std, with_coslat, with_weights",
    [
        (True, True, True),
        (True, False, True),
        (False, True, True),
        (False, False, True),
        (True, True, False),
        (True, False, False),
        (False, True, False),
        (False, False, False),
    ],
)
def test_transform_params(with_std, with_coslat, with_weights, mock_data_array):
    s = Scaler(with_std=with_std, with_coslat=with_coslat)
    sample_dims = ["time"]
    feature_dims = ["lat", "lon"]
    size_lats = mock_data_array.lat.size
    if with_weights:
        weights = xr.DataArray(
            np.random.rand(size_lats), dims=["lat"], coords={"lat": mock_data_array.lat}
        )
    else:
        weights = None
    s.fit(mock_data_array, sample_dims, feature_dims, weights)
    transformed = s.transform(mock_data_array)
    assert transformed is not None, "Transformed data is None."

    transformed_mean = transformed.mean(sample_dims, skipna=False)
    assert np.allclose(transformed_mean, 0), "Mean of the transformed data is not zero."

    if with_std and not (with_coslat or with_weights):
        transformed_std = transformed.std(sample_dims, skipna=False)

        assert np.allclose(
            transformed_std, 1
        ), "Standard deviation of the transformed data is not one."

    if with_coslat:
        assert s.coslat_weights_ is not None, "Scaler coslat_weights is None."
        assert not np.array_equal(
            transformed, mock_data_array
        ), "Data has not been transformed."

    transformed2 = s.fit_transform(mock_data_array, sample_dims, feature_dims, weights)
    xr.testing.assert_allclose(transformed, transformed2)


@pytest.mark.parametrize(
    "with_std, with_coslat",
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_inverse_transform_params(with_std, with_coslat, mock_data_array):
    s = Scaler(
        with_std=with_std,
        with_coslat=with_coslat,
    )
    sample_dims = ["time"]
    feature_dims = ["lat", "lon"]
    size_lats = mock_data_array.lat.size
    weights = xr.DataArray(
        np.random.rand(size_lats), dims=["lat"], coords={"lat": mock_data_array.lat}
    )
    s.fit(mock_data_array, sample_dims, feature_dims, weights)
    transformed = s.transform(mock_data_array)
    inverted = s.inverse_transform_data(transformed)
    xr.testing.assert_allclose(inverted, mock_data_array)


@pytest.mark.parametrize(
    "dim_sample, dim_feature",
    [
        (("time",), ("lat", "lon")),
        (("time",), ("lon", "lat")),
        (("lat", "lon"), ("time",)),
        (("lon", "lat"), ("time",)),
    ],
)
def test_fit_dims(dim_sample, dim_feature, mock_data_array):
    s = Scaler(with_std=True)
    s.fit(mock_data_array, dim_sample, dim_feature)
    assert hasattr(s, "mean_"), "Scaler has no mean attribute."
    assert s.mean_ is not None, "Scaler mean is None."
    assert hasattr(s, "std_"), "Scaler has no std attribute."
    assert s.std_ is not None, "Scaler std is None."
    # check that all dimensions are present except the sample dimensions
    assert set(s.mean_.dims) == set(mock_data_array.dims) - set(
        dim_sample
    ), "Mean has wrong dimensions."
    assert set(s.std_.dims) == set(mock_data_array.dims) - set(
        dim_sample
    ), "Standard deviation has wrong dimensions."


@pytest.mark.parametrize(
    "dim_sample, dim_feature",
    [
        (("time",), ("lat", "lon")),
        (("time",), ("lon", "lat")),
        (("lat", "lon"), ("time",)),
        (("lon", "lat"), ("time",)),
    ],
)
def test_fit_transform_dims(dim_sample, dim_feature, mock_data_array):
    s = Scaler()
    transformed = s.fit_transform(mock_data_array, dim_sample, dim_feature)
    # check that all dimensions are present
    assert set(transformed.dims) == set(
        mock_data_array.dims
    ), "Transformed data has wrong dimensions."
    # check that the coordinates are the same
    for dim in mock_data_array.dims:
        xr.testing.assert_allclose(transformed[dim], mock_data_array[dim])


# Test input types
def test_fit_input_type(mock_data_array, mock_dataset, mock_data_array_list):
    s = Scaler()

    with pytest.raises(TypeError):
        s.fit(mock_data_array_list, ["time"], ["lon", "lat"])

    s.fit(mock_data_array, ["time"], ["lon", "lat"])

    with pytest.raises(TypeError):
        s.transform(mock_data_array_list)



from xeofs import cross, multi, single, validation
from xeofs._version import __version__
from xeofs.rotator_factory import RotatorFactory

__all__ = ["single", "cross", "multi", "RotatorFactory", "validation", "__version__"]



import pytest
import xarray as xr

from xeofs.data_container import DataContainer
from xeofs.single import HilbertEOF, HilbertEOFRotator


@pytest.fixture
def ceof_model(mock_data_array, dim):
    ceof = HilbertEOF(n_modes=5)
    ceof.fit(mock_data_array, dim)
    return ceof


@pytest.fixture
def ceof_model_delayed(mock_dask_data_array, dim):
    ceof = HilbertEOF(n_modes=5)
    ceof.fit(mock_dask_data_array, dim)
    return ceof


def test_init():
    # Instantiate the HilbertEOFRotator class
    ceof_rotator = HilbertEOFRotator(n_modes=3, power=2, max_iter=100, rtol=1e-6)

    assert ceof_rotator._params["n_modes"] == 3
    assert ceof_rotator._params["power"] == 2
    assert ceof_rotator._params["max_iter"] == 100
    assert ceof_rotator._params["rtol"] == 1e-6


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(ceof_model):
    ceof_rotator = HilbertEOFRotator(n_modes=3)
    ceof_rotator.fit(ceof_model)

    assert hasattr(
        ceof_rotator, "model_data"
    ), 'The attribute "model_data" should be populated after fitting.'
    assert hasattr(
        ceof_rotator, "data"
    ), 'The attribute "data" should be populated after fitting.'
    assert isinstance(ceof_rotator.model_data, DataContainer)
    assert isinstance(ceof_rotator.data, DataContainer)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform_not_implemented(ceof_model, mock_data_array):
    ceof_rotator = HilbertEOFRotator(n_modes=3)
    ceof_rotator.fit(ceof_model)

    with pytest.raises(NotImplementedError):
        ceof_rotator.transform(mock_data_array)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_inverse_transform(ceof_model):
    ceof_rotator = HilbertEOFRotator(n_modes=3)
    ceof_rotator.fit(ceof_model)
    scores = ceof_rotator.data["scores"].isel(mode=1)
    Xrec = ceof_rotator.inverse_transform(scores)

    assert isinstance(Xrec, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_amplitude(ceof_model):
    ceof_rotator = HilbertEOFRotator(n_modes=3)
    ceof_rotator.fit(ceof_model)
    comps_amp = ceof_rotator.components_amplitude()

    assert isinstance(comps_amp, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_phase(ceof_model):
    ceof_rotator = HilbertEOFRotator(n_modes=3)
    ceof_rotator.fit(ceof_model)
    comps_phase = ceof_rotator.components_phase()

    assert isinstance(comps_phase, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_amplitude(ceof_model):
    ceof_rotator = HilbertEOFRotator(n_modes=3)
    ceof_rotator.fit(ceof_model)
    scores_amp = ceof_rotator.scores_amplitude()

    assert isinstance(scores_amp, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_phase(ceof_model):
    ceof_rotator = HilbertEOFRotator(n_modes=3)
    ceof_rotator.fit(ceof_model)
    scores_phase = ceof_rotator.scores_phase()

    assert isinstance(scores_phase, xr.DataArray)



import warnings

import dask
import numpy as np
import xarray as xr
from dask.array import Array as DaskArray  # type: ignore
from dask.array.linalg import svd_compressed as dask_svd
from scipy.sparse.linalg import svds as complex_svd  # type: ignore
from sklearn.utils.extmath import randomized_svd

from ..utils.sanity_checks import sanity_check_n_modes
from ..utils.xarray_utils import get_deterministic_sign_multiplier


class Decomposer:
    """Decomposes a data object using Singular Value Decomposition (SVD).

    The data object will be decomposed like X = U * S * V.T, where U and V are
    orthogonal matrices and S is a diagonal matrix with the singular values on
    the diagonal.

    Parameters
    ----------
    n_modes : int | float
        Number of components to be computed. If float, it represents the fraction of variance to keep.
    init_rank_reduction : float, default=0.0
        Used only when `n_modes` is given as a float. Specifiy the initial target rank to be computed by randomized SVD before truncating the solution to the desired fraction of explained variance. Must be in the half open interval (0, 1]. Lower values will speed up the computation. If the rank is too low and the fraction of explained variance is not reached, a warning will be raised.
    flip_signs : bool, default=True
        Whether to flip the sign of the components to ensure deterministic output.
    compute : bool, default=True
        Whether to compute the decomposition immediately.
    solver : {'auto', 'full', 'randomized'}, default='auto'
        The solver is selected by a default policy based on size of `X` and `n_modes`:
        if the input data is larger than 500x500 and the number of modes to extract is lower
        than 80% of the smallest dimension of the data, then the more efficient
        `randomized` method is enabled. Otherwise the exact full SVD is computed
        and optionally truncated afterwards.
    random_state : np.random.Generator | int | None, default=None
        Seed for the random number generator.
    component_dim_name : str, default='mode'
        Name of the component dimension in the output DataArrays.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver.
    """

    def __init__(
        self,
        n_modes: int | float,
        init_rank_reduction: float = 0.3,
        flip_signs: bool = True,
        compute: bool = True,
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        component_dim_name: str = "mode",
        solver_kwargs: dict = {},
    ):
        sanity_check_n_modes(n_modes)
        self.is_based_on_variance = False if isinstance(n_modes, int) else True

        if self.is_based_on_variance:
            if not (0 < init_rank_reduction <= 1.0):
                raise ValueError(
                    "init_rank_reduction must be in the half open interval (0, 1]."
                )

        self.n_modes = n_modes
        self.n_modes_precompute = n_modes
        self.init_rank_reduction = init_rank_reduction
        self.flip_signs = flip_signs
        self.compute = compute
        self.solver = solver
        self.random_state = random_state
        self.component_dim_name = component_dim_name
        self.solver_kwargs = solver_kwargs

    def fit(self, X, dims=("sample", "feature")):
        """Decomposes the data object.

        Parameters
        ----------
        X : DataArray
            A 2-dimensional data object to be decomposed.
        dims : tuple of str
            Dimensions of the data object.
        """
        rank = min(X.shape)

        if self.is_based_on_variance:
            self.n_modes_precompute = int(rank * self.init_rank_reduction)
            if self.n_modes_precompute < 1:
                warnings.warn(
                    f"`init_rank_reduction={self.init_rank_reduction}` is too low resulting in zero components. One component will be computed instead."
                )
                self.n_modes_precompute = 1

        # TODO(nicrie): perhaps we can just set n_modes to rank if it is larger than rank (possible solution for #158)
        if self.n_modes_precompute > rank:
            raise ValueError(
                f"n_modes must be less than or equal to the rank of the dataset (rank = {rank})."
            )

        # Check if data is small enough to use exact SVD
        # If not, use randomized SVD
        # If data is complex, use scipy sparse SVD
        # If data is dask, use dask SVD
        # Conditions for using exact SVD follow scitkit-learn's PCA implementation
        # Source: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html

        use_dask = True if isinstance(X.data, DaskArray) else False
        use_complex = True if np.iscomplexobj(X.data) else False

        is_small_data = max(X.shape) < 500

        match self.solver:
            # TODO(nicrie): implement more performant case for tall and skinny problems which are best handled by precomputing the covariance matrix.
            # if X.shape[1] <= 1_000 and X.shape[0] >= 10 * X.shape[1]: -> covariance_eigh" (see sklearn PCA implementation: https://github.com/scikit-learn/scikit-learn/blob/e87b32a81c70abed8f2e97483758eb64df8255e9/sklearn/decomposition/_pca.py#L526)
            case "auto":
                use_exact = (
                    True
                    if is_small_data and self.n_modes_precompute > int(0.8 * rank)
                    else False
                )
            case "full":
                use_exact = True
            case "randomized":
                use_exact = False
            case _:
                raise ValueError(
                    f"Unrecognized solver '{self.solver}'. "
                    "Valid options are 'auto', 'full', and 'randomized'."
                )

        # Use exact SVD for small data sets
        if use_exact:
            U, s, VT = self._svd(X, dims, np.linalg.svd, self.solver_kwargs)
            U = U[:, : self.n_modes_precompute]
            s = s[: self.n_modes_precompute]
            VT = VT[: self.n_modes_precompute, :]

        # Use randomized SVD for large, real-valued data sets
        elif (not use_complex) and (not use_dask):
            solver_kwargs = self.solver_kwargs | {
                "n_components": self.n_modes_precompute,
                "random_state": self.random_state,
            }
            U, s, VT = self._svd(X, dims, randomized_svd, solver_kwargs)

        # Use scipy sparse SVD for large, complex-valued data sets
        elif use_complex and (not use_dask):
            # Scipy sparse version
            solver_kwargs = self.solver_kwargs | {
                "k": self.n_modes_precompute,
                "solver": "lobpcg",
                "random_state": self.random_state,
            }
            U, s, VT = self._svd(X, dims, complex_svd, solver_kwargs)
            idx_sort = np.argsort(s)[::-1]
            U = U[:, idx_sort]
            s = s[idx_sort]
            VT = VT[idx_sort, :]

        # Use dask SVD for large, real-valued, delayed data sets
        elif (not use_complex) and use_dask:
            solver_kwargs = self.solver_kwargs | {
                "k": self.n_modes_precompute,
                "seed": self.random_state,
            }
            solver_kwargs.setdefault("compute", self.compute)
            solver_kwargs.setdefault("n_power_iter", 4)
            U, s, VT = self._svd(X, dims, dask_svd, solver_kwargs)
            U, s, VT = self._compute_svd_result(U, s, VT)
        else:
            err_msg = (
                "Complex data together with dask is currently not implemented. See dask issue 7639 "
                "https://github.com/dask/dask/issues/7639"
            )
            raise NotImplementedError(err_msg)

        U = U.assign_coords(mode=range(1, U.mode.size + 1))
        s = s.assign_coords(mode=range(1, U.mode.size + 1))
        VT = VT.assign_coords(mode=range(1, U.mode.size + 1))

        U.name = "U"
        s.name = "s"
        VT.name = "V"

        # Truncate the decomposition to the desired number of modes
        if self.is_based_on_variance:
            # Truncating based on variance requires computation of dask array
            # which we prefer to avoid
            if use_dask:
                err_msg = "Estimating the number of modes to keep based on variance is not supported with dask arrays. Please explicitly specifiy the number of modes to keep by using an integer for the number of modes."
                raise ValueError(err_msg)

            # Compute the fraction of explained variance per mode
            N = X.shape[0] - 1
            total_variance = X.var(X.dims[0], ddof=1).sum(X.dims[1])
            explained_variance = s**2 / N / total_variance
            cum_expvar = explained_variance.cumsum(self.component_dim_name)
            total_explained_variance = cum_expvar[-1].item()

            n_modes_required = (
                self.n_modes_precompute
                - (cum_expvar >= self.n_modes).sum(self.component_dim_name)
                + 1
            )
            if n_modes_required > self.n_modes_precompute:
                warnings.warn(
                    f"Dataset has {self.n_modes_precompute} components, explaining {total_explained_variance:.2%} of the variance. However, {self.n_modes:.2%} explained variance was requested. Please consider increasing `init_rank_reduction`."
                )
                n_modes_required = self.n_modes_precompute

            # Truncate solution to the desired fraction of explained variance
            U = U.sel(mode=slice(1, n_modes_required))
            s = s.sel(mode=slice(1, n_modes_required))
            VT = VT.sel(mode=slice(1, n_modes_required))

        # Flip signs of components to ensure deterministic output
        if self.flip_signs:
            sign_multiplier = get_deterministic_sign_multiplier(VT, dims[1])
            VT *= sign_multiplier
            U *= sign_multiplier

        self.U_ = U
        self.s_ = s
        self.V_ = VT.conj().transpose(dims[1], self.component_dim_name)

    def _svd(self, X, dims, func, kwargs):
        """Performs SVD on the data

        Parameters
        ----------
        X : DataArray
            A 2-dimensional data object to be decomposed.
        dims : tuple of str
            Dimensions of the data object.
        func : Callable
            Method to perform SVD.
        kwargs : dict
            Additional keyword arguments passed to the SVD solver.

        Returns
        -------
        U : DataArray
            Left singular vectors.
        s : DataArray
            Singular values.
        VT : DataArray
            Right singular vectors.
        """
        try:
            U, s, VT = xr.apply_ufunc(
                func,
                X,
                kwargs=kwargs,
                input_core_dims=[dims],
                output_core_dims=[
                    [dims[0], self.component_dim_name],
                    [self.component_dim_name],
                    [self.component_dim_name, dims[1]],
                ],
                dask="allowed",
            )
            return U, s, VT
        except np.linalg.LinAlgError:
            raise np.linalg.LinAlgError(
                "SVD failed. This may be due to isolated NaN values in the data. Please consider the following steps:\n"
                "1. Check for and remove any isolated NaNs in your dataset.\n"
                "2. If the error persists, please raise an issue at https://github.com/xarray-contrib/xeofs/issues."
            )

    def _compute_svd_result(self, U, s, VT):
        """Computes the SVD result.

        Parameters
        ----------
        U : DataArray
            Left singular vectors.
        s : DataArray
            Singular values.
        VT : DataArray
            Right singular vectors.

        Returns
        -------
        U : DataArray
            Left singular vectors.
        s : DataArray
            Singular values.
        VT : DataArray
            Right singular vectors.
        """
        match self.compute:
            case False:
                pass
            case True:
                U, s, VT = dask.compute(U, s, VT)
        return U, s, VT



import numpy as np
import pytest
import xarray as xr

from xeofs.single import EOF
from xeofs.validation import EOFBootstrapper


# If the data is defined as a fixture (e.g., mock_data_array),
# we can't directly parametrize it with the @pytest.mark.parametrize decorator.
# A way to go around this is to create a fixture that uses another fixture.
# This might be a bit complex, but it should work
@pytest.fixture(params=["mock_data_array", "mock_dataset", "mock_data_array_list"])
def data(request):
    # Here, request.param would be a fixture name as a string
    return request.getfixturevalue(request.param)


@pytest.fixture
def dim(request):
    return request.param


@pytest.fixture
def eof_model(data, dim):
    eof = EOF()
    eof.fit(data, dim)
    return eof


def test_init():
    bootstrapper = EOFBootstrapper(n_bootstraps=3, seed=7)
    assert bootstrapper._params["n_bootstraps"] == 3
    assert bootstrapper._params["seed"] == 7


@pytest.mark.parametrize(
    "data, dim",
    [
        ("mock_data_array", "time"),
        ("mock_data_array", ("lon", "lat")),
        ("mock_data_array", ("lat", "lon")),
        ("mock_dataset", "time"),
        ("mock_dataset", ("lon", "lat")),
        ("mock_dataset", ("lat", "lon")),
        ("mock_data_array_list", "time"),
        ("mock_data_array_list", ("lon", "lat")),
        ("mock_data_array_list", ("lat", "lon")),
    ],
    indirect=["data", "dim"],
)
def test_fit(eof_model):
    """Bootstrapping creates DataArrays with expected dims and coords"""
    bootstrapper = EOFBootstrapper(n_bootstraps=3, seed=7)
    bootstrapper.fit(eof_model)

    # DataArrays are created
    assert isinstance(
        bootstrapper.data["explained_variance"], xr.DataArray
    ), "explained variance is not a DataArray"
    assert isinstance(
        bootstrapper.data["components"], xr.DataArray
    ), "components is not a DataArray"
    assert isinstance(
        bootstrapper.data["scores"], xr.DataArray
    ), "scores is not a DataArray"

    # DataArrays have expected dims
    expected_dims = set(eof_model.data["explained_variance"].dims)
    expected_dims.add("n")
    true_dims = set(bootstrapper.data["explained_variance"].dims)
    err_message = (
        f"explained variance dimensions are {true_dims} instead of {expected_dims}"
    )
    assert true_dims == expected_dims, err_message

    expected_dims = set(eof_model.data["components"].dims)
    expected_dims.add("n")
    true_dims = set(bootstrapper.data["components"].dims)
    err_message = f"components dimensions are {true_dims} instead of {expected_dims}"
    assert true_dims == expected_dims, err_message

    expected_dims = set(eof_model.data["scores"].dims)
    expected_dims.add("n")
    true_dims = set(bootstrapper.data["scores"].dims)
    err_message = f"scores dimensions are {true_dims} instead of {expected_dims}"
    assert true_dims == expected_dims, err_message

    # DataArrays have expected coords
    ref_da = eof_model.data["explained_variance"]
    test_da = bootstrapper.data["explained_variance"]
    for dim, coords in ref_da.coords.items():
        assert test_da[dim].equals(
            coords
        ), f"explained variance coords for {dim} are not equal"

    ref_da = eof_model.data["components"]
    test_da = bootstrapper.data["components"]
    for dim, coords in ref_da.coords.items():
        assert test_da[dim].equals(coords), f"components coords for {dim} are not equal"

    ref_da = eof_model.data["scores"]
    test_da = bootstrapper.data["scores"]
    for dim, coords in ref_da.coords.items():
        assert test_da[dim].equals(coords), f"scores coords for {dim} are not equal"


@pytest.mark.parametrize(
    "data, dim",
    [
        ("mock_data_array", "time"),
        ("mock_data_array", ("lon", "lat")),
        ("mock_data_array", ("lat", "lon")),
        ("mock_dataset", "time"),
        ("mock_dataset", ("lon", "lat")),
        ("mock_dataset", ("lat", "lon")),
        ("mock_data_array_list", "time"),
        ("mock_data_array_list", ("lon", "lat")),
        ("mock_data_array_list", ("lat", "lon")),
    ],
    indirect=["data", "dim"],
)
def test_seed(eof_model):
    """Bootstrapping creates DataArrays with expected dims and coords"""
    bootstrapper1 = EOFBootstrapper(n_bootstraps=5, seed=7)
    bootstrapper1.fit(eof_model)
    expvar1 = bootstrapper1.explained_variance()

    bootstrapper2 = EOFBootstrapper(n_bootstraps=5, seed=7)
    bootstrapper2.fit(eof_model)
    expvar2 = bootstrapper2.explained_variance()

    assert np.allclose(
        expvar1.values, expvar2.values
    ), "bootstrapping with the same seed does not produce the same results"


@pytest.mark.parametrize(
    "data, dim",
    [
        ("mock_data_array", "time"),
        ("mock_data_array", ("lon", "lat")),
        ("mock_data_array", ("lat", "lon")),
        ("mock_dataset", "time"),
        ("mock_dataset", ("lon", "lat")),
        ("mock_dataset", ("lat", "lon")),
        ("mock_data_array_list", "time"),
        ("mock_data_array_list", ("lon", "lat")),
        ("mock_data_array_list", ("lat", "lon")),
    ],
    indirect=["data", "dim"],
)
def test_explained_variance(eof_model):
    """Bootstrapping creates DataArrays expected dims and coords"""
    bootstrapper = EOFBootstrapper(n_bootstraps=3)
    bootstrapper.fit(eof_model)

    expvar = bootstrapper.explained_variance()
    assert isinstance(expvar, xr.DataArray), "explained variance is not a DataArray"

    # DataArrays have expected dims
    ref = eof_model.explained_variance()
    expected_dims = set(ref.dims)
    expected_dims.add("n")
    true_dims = set(expvar.dims)
    err_message = (
        f"explained variance dimensions are {true_dims} instead of {expected_dims}"
    )
    assert true_dims == expected_dims, err_message

    # DataArrays have expected coords
    for dim, coords in ref.coords.items():
        assert expvar[dim].equals(
            coords
        ), f"explained variance coords for {dim} are not equal"


@pytest.mark.parametrize(
    "data, dim",
    [
        ("mock_data_array", "time"),
        ("mock_data_array", ("lon", "lat")),
        ("mock_data_array", ("lat", "lon")),
        ("mock_dataset", "time"),
        ("mock_dataset", ("lon", "lat")),
        ("mock_dataset", ("lat", "lon")),
        ("mock_data_array_list", "time"),
        ("mock_data_array_list", ("lon", "lat")),
        ("mock_data_array_list", ("lat", "lon")),
    ],
    indirect=["data", "dim"],
)
def test_explained_variance_ratio(eof_model):
    """Bootstrapping creates DataArrays expected dims and coords"""
    bootstrapper = EOFBootstrapper(n_bootstraps=3)
    bootstrapper.fit(eof_model)

    expvar = bootstrapper.explained_variance_ratio()
    assert isinstance(
        expvar, xr.DataArray
    ), "explained variance ratio is not a DataArray"

    # DataArrays have expected dims
    ref = eof_model.explained_variance_ratio()
    expected_dims = set(ref.dims)
    expected_dims.add("n")
    true_dims = set(expvar.dims)
    err_message = f"explained variance ratio dimensions are {true_dims} instead of {expected_dims}"
    assert true_dims == expected_dims, err_message

    # DataArrays have expected coords
    for dim, coords in ref.coords.items():
        assert expvar[dim].equals(
            coords
        ), f"explained variance ratio coords for {dim} are not equal"


@pytest.mark.parametrize(
    "data, dim, expected_type",
    [
        ("mock_data_array", "time", xr.DataArray),
        ("mock_data_array", ("lon", "lat"), xr.DataArray),
        ("mock_data_array", ("lat", "lon"), xr.DataArray),
        ("mock_dataset", "time", xr.Dataset),
        ("mock_dataset", ("lon", "lat"), xr.Dataset),
        ("mock_dataset", ("lat", "lon"), xr.Dataset),
    ],
    indirect=["data", "dim"],
)
def test_components(eof_model, expected_type):
    """Bootstrapping components creates expected dims and coords"""
    bootstrapper = EOFBootstrapper(n_bootstraps=3)
    bootstrapper.fit(eof_model)

    components = bootstrapper.components()
    assert isinstance(components, expected_type), f"components is not a {expected_type}"

    # check for expected dimensions
    ref = eof_model.components()
    expected_dims = set(ref.dims)
    expected_dims.add("n")
    true_dims = set(components.dims)
    err_message = f"components dimensions are {true_dims} instead of {expected_dims}"
    assert true_dims == expected_dims, err_message

    # check for expected coords
    for dim, coords in ref.coords.items():
        assert components[dim].equals(
            coords
        ), f"components coords for {dim} are not equal"


@pytest.mark.parametrize(
    "data, dim, expected_type",
    [
        ("mock_data_array_list", "time", list),
        ("mock_data_array_list", ("lon", "lat"), list),
        ("mock_data_array_list", ("lat", "lon"), list),
    ],
    indirect=["data", "dim"],
)
def test_components_list(eof_model, expected_type):
    """Bootstrapping components creates expected dims and coords"""
    bootstrapper = EOFBootstrapper(n_bootstraps=3)
    bootstrapper.fit(eof_model)

    comps_list = bootstrapper.components()
    assert isinstance(comps_list, expected_type), f"components is not a {expected_type}"

    # check for expected dimensions
    ref_list = eof_model.components()
    for ref, comps in zip(ref_list, comps_list):
        expected_dims = set(ref.dims)
        expected_dims.add("n")
        true_dims = set(comps.dims)
        err_message = (
            f"components dimensions are {true_dims} instead of {expected_dims}"
        )
        assert true_dims == expected_dims, err_message

    # check for expected coords
    for ref, comps in zip(ref_list, comps_list):
        for dim, coords in ref.coords.items():
            assert comps[dim].equals(
                coords
            ), f"components coords for {dim} are not equal"


@pytest.mark.parametrize(
    "data, dim",
    [
        ("mock_data_array", "time"),
        ("mock_data_array", ("lon", "lat")),
        ("mock_data_array", ("lat", "lon")),
        ("mock_dataset", "time"),
        ("mock_dataset", ("lon", "lat")),
        ("mock_dataset", ("lat", "lon")),
        ("mock_data_array_list", "time"),
        ("mock_data_array_list", ("lon", "lat")),
        ("mock_data_array_list", ("lat", "lon")),
    ],
    indirect=["data", "dim"],
)
def test_scores(eof_model):
    """Bootstrapping scores creates expected dims and coords"""
    bootstrapper = EOFBootstrapper(n_bootstraps=3)
    bootstrapper.fit(eof_model)

    scores = bootstrapper.scores()
    assert isinstance(scores, xr.DataArray), "scores is not a xr.DataArray"

    # check for expected dimensions
    ref = eof_model.scores()
    expected_dims = set(ref.dims)
    expected_dims.add("n")
    true_dims = set(scores.dims)
    err_message = f"scores dimensions are {true_dims} instead of {expected_dims}"
    assert true_dims == expected_dims, err_message

    # check for expected coords
    for dim, coords in ref.coords.items():
        assert scores[dim].equals(coords), f"scores coords for {dim} are not equal"



import pytest

from xeofs.preprocessing.multi_index_converter import (
    MultiIndexConverter,
)
from ..utilities import assert_expected_dims, data_is_dask, data_has_multiindex

# =============================================================================
# GENERALLY VALID TEST CASES
# =============================================================================
N_VARIABLES = [1, 2]
N_SAMPLE_DIMS = [1, 2]
N_FEATURE_DIMS = [1, 2]
INDEX_POLICY = ["index"]
NAN_POLICY = ["no_nan"]
DASK_POLICY = ["no_dask", "dask"]
SEED = [0]

VALID_TEST_DATA = [
    (nv, ns, nf, index, nan, dask)
    for nv in N_VARIABLES
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in NAN_POLICY
    for dask in DASK_POLICY
]


# TESTS
# =============================================================================
@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_transform(synthetic_dataset):
    converter = MultiIndexConverter()
    converter.fit(synthetic_dataset)
    transformed_data = converter.transform(synthetic_dataset)

    is_dask_before = data_is_dask(synthetic_dataset)
    is_dask_after = data_is_dask(transformed_data)

    # Transforming does not affect dimensions
    assert_expected_dims(transformed_data, synthetic_dataset, policy="all")

    # Transforming doesn't change the dask-ness of the data
    assert is_dask_before == is_dask_after

    # Transforming removes MultiIndex
    assert data_has_multiindex(transformed_data) is False

    # Result is robust to calling the method multiple times
    transformed_data = converter.transform(synthetic_dataset)
    assert data_has_multiindex(transformed_data) is False

    # Transforming data twice won't change the data
    transformed_data2 = converter.transform(transformed_data)
    assert data_has_multiindex(transformed_data2) is False
    assert transformed_data.identical(transformed_data2)


@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_inverse_transform(synthetic_dataset):
    converter = MultiIndexConverter()
    converter.fit(synthetic_dataset)
    transformed_data = converter.transform(synthetic_dataset)
    inverse_transformed_data = converter.inverse_transform_data(transformed_data)

    is_dask_before = data_is_dask(synthetic_dataset)
    is_dask_after = data_is_dask(transformed_data)

    # Transforming doesn't change the dask-ness of the data
    assert is_dask_before == is_dask_after

    has_multiindex_before = data_has_multiindex(synthetic_dataset)
    has_multiindex_after = data_has_multiindex(inverse_transformed_data)

    assert inverse_transformed_data.identical(synthetic_dataset)
    assert has_multiindex_before == has_multiindex_after
